/***************************************************************************
 *   Copyright (C) 2010 by Ahsan Tariq   *
 *   ahsan@mpi-sws.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <iostream>
#include <vector>
#include <string>
#include <cstring>
#include <fstream>
#include <cstdlib>

#include "Algorithms.h"
#include "RIB.h"


using namespace std;

int inter = 0;
int cSMALTA::GetMask (std::vector<int>& ImmPref, int d)	{
  for (int i = d; i >= 0; i--)	
    if (ImmPref[i] == 1)	
      return (d - i);
  return -1;
}


int cSMALTA::ApplyORTCNR ()	{
  int ret;
  ReadRIB();
  ORTCNRPass1();
  ORTCNRPass2();
  ret = ORTCNRPass3();
  return ret;
}

int cSMALTA::ConvertNHtoIFace (unsigned int NH)	{
  for (int i = 0; i < (int)NHtoIFaceMap.size(); i++)	{
    //std::cout << "iterating " <<  NHtoIFaceMap[i] << "\n";
    if (NH == NHtoIFaceMap[i])	{
      //cout << NHtoIFaceMap [i];
      return i;
    }
  }
  return -2;
}



/*************************************************************************************************************************
Read the RIB and map the unique nexthops to self defined interface numbers. 
ORTC is applied only on INFIB entries. If ORTC is running for the first time it converts all non-null NH to INFIB
Additionaly construct the temporary datastructure in a one to one map with the RIBNodes
*************************************************************************************************************************/

int cSMALTA::ReadRIB ()	{
  cRIBNode *next = RIB->root,*r;
  cTempORTCNRData *ORTCNRDataTree = TempORTCNRData, *t;
  cTempORTCNRData *newData;
  std::vector <int> flags(33,0);
  int depth = 0;
  int match = 0;
  // introduce DROP nexthop at the root
  (*next).Data.Routes.BestNH = DROP;
  (*next).Data.Routes.BestASLen = 0;
  // set it as original	
  (*next).Data.SMALTAFlags.Original = 1;	
  (*next).Data.RIBFlags.INFIB = 0;
  int iter = 0;
  int nodes = 0;
  while (next != NULL)	{
    switch (flags[depth])	{
    case 0:						// first time visit to a node
      // search the NHtoIFaceMap if this NH exists
      nodes++;
      if (FirstTimeORTC == 0)
	(*next).Data.SMALTAFlags.Original = 0;
      if ((*next).Data.Routes.BestNH != NULL_NEXTHOP)	{

	if (FirstTimeORTC == 0)	{		//First time application of ORTC
						//std::cout << "First time ORTC\n";
						//FirstTimeORTC = 1;
	  (*next).Data.RIBFlags.INFIB = 0;		
	  (*next).Data.SMALTAFlags.DistanceToParent = -1;
	  //if ((*next).Data.Routes.BestNH != DROP)
	  if ((*next).Data.Routes.BestNH != NULL_NEXTHOP)	{
	    (*next).Data.SMALTAFlags.Original = 1;
	    iter++;
	  }
	  //FirstTimeORTC = 1;
	}						
	else	{
	  iter++;
	  if ((*next).Data.SMALTAFlags.Original == 0)	{
	    (*next).Data.RIBFlags.INFIB = 0;
	    iter--;
	    (*next).Data.SMALTAFlags.DistanceToParent = -1;
	    if ((*next).parent !=NULL)	{
	      (*next).Data.Routes.BestNH = NULL_NEXTHOP;
	      (*next).Data.Routes.BestASLen = 0;
	    }
	    if (next->child[LEFT] == NULL && next->child[RIGHT] == NULL)	{
	      nodes--;
	      r = next;
	      next = next->parent;
	      t = ORTCNRDataTree;
	      ORTCNRDataTree = ORTCNRDataTree->parent;
	      depth--;
	      if ((*next).child[LEFT] == r)	{
		(*next).child[LEFT] = NULL;
		delete r;
		ORTCNRDataTree->child[LEFT] = NULL;
		delete t;
	      }
	      else	{
		(*next).child[RIGHT] = NULL;
		delete r;
		ORTCNRDataTree->child[RIGHT] = NULL;
		delete t;
	      }
	      break;
	    }
	  }
	}
	if ((*next).Data.Routes.BestNH != NULL_NEXTHOP)	{
	  for (unsigned int i = 0; i < NHtoIFaceMap.size(); i++)	{
	    if ((*next).Data.Routes.BestNH == NHtoIFaceMap[i])	{
	      match = 1;
	      //std::cout << "match " << (*next).Data.Routes.BestNH << "\n";
	    }
	  }
	  if (!match)	{	
	    //cout << "adding " << (*next).Data.Routes.BestNH << "\n";
	    NHtoIFaceMap.push_back ((*next).Data.Routes.BestNH);
	  }
					
					
	  (*ORTCNRDataTree).IFace = ConvertNHtoIFace ((*next).Data.Routes.BestNH);
	  if ((*ORTCNRDataTree).IFace == -2)	{
	    std::cout << "Error in converting NH " << (*next).Data.Routes.BestNH <<  "to Interface. Exiting... \n";	
	    exit(1);
	  }
					
	  match = 0;
	}	
      }
      flags[depth] = 1;
      if ((*next).child[LEFT] == NULL)	{
	flags[depth] = 2;
	if ((*next).child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  r = next;
	  next = (*next).parent;
	  t = ORTCNRDataTree;
	  depth--;
	  ORTCNRDataTree = (*ORTCNRDataTree).parent;
	}
	else	{
	  next = (*next).child[RIGHT];
	  depth++;
	  newData = new cTempORTCNRData (NULL, ORTCNRDataTree, NULL, NULL, -1, -1);	
	  (*ORTCNRDataTree).child[RIGHT] = newData;
	  ORTCNRDataTree = newData;
	}
      }
      else	{
	next = (*next).child[LEFT];
	depth++;
	newData = new cTempORTCNRData (NULL, ORTCNRDataTree, NULL, NULL, -1, -1);	
	(*ORTCNRDataTree).child[LEFT] = newData;
	ORTCNRDataTree = newData;
      }
      break;
			
    case 1:
      flags[depth] = 2;
      if ((*next).child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = (*next).parent;
	depth--;
	ORTCNRDataTree = (*ORTCNRDataTree).parent;
      }
      else	{
	next = (*next).child[RIGHT];
	depth++;
	newData = new cTempORTCNRData (NULL, ORTCNRDataTree, NULL, NULL, -1, -1);	
	(*ORTCNRDataTree).child[RIGHT] = newData;
	ORTCNRDataTree = newData;
      }
      break;
			
    case 2:
      flags[depth] = 0;
      next = (*next).parent;
      depth--;
      ORTCNRDataTree = (*ORTCNRDataTree).parent;
      break;
    }	// end switch
  }	// end while
  FirstTimeORTC = 1;
  //std::cout << "Total Originals = " << iter << "\n";
  //cout << "Total Nodes = " << nodes << endl;
  return 0;
}

int cSMALTA::ORTCNRPass1 ()	{
  //std::cout << "ORTC Pass 1\n";
  cTempORTCNRData * ORTCData = TempORTCNRData, *newORTC;
  cRIBNode *next = RIB->root, *newRIBNode;
  int depth = 0;
  std::vector <int> flags (33,0);
  unsigned int prefix = 0;
  std::vector <int> ImmediatePrefix (33,0);
  leaves = 0;
  int newleaves = 0;
  while (ORTCData != NULL)	{
    switch (flags[depth])	{
			
    case 0:		
      //std::cout << ORTCData->IFace << "\n";
      if (ORTCData->IFace == -1)
	ORTCData->IFace = ORTCData->parent->IFace;
      else
	ImmediatePrefix[depth] = 1;
      /* Assign every leaf node the mask offset from its immediate valued parent */
      if (ORTCData->child[LEFT] == NULL && ORTCData->child[RIGHT] != NULL)	{		
	newORTC = 
	  new cTempORTCNRData (NULL, ORTCData,NULL, NULL, 
			       -1, (GetMask (ImmediatePrefix,depth)+1));
	ORTCData->child[LEFT] = newORTC;
	newRIBNode = 
	  new cRIBNode (0, 0,next, NULL, NULL,RIB->pol);
	next->child[LEFT] = newRIBNode;	
	(*newRIBNode).Data.SMALTAFlags.DistanceToParent = -1;
	(*newRIBNode).Data.SMALTAFlags.Original = 0;
	newleaves++;
      }
      if (ORTCData->child[LEFT] != NULL && ORTCData->child[RIGHT] == NULL)	{
	newORTC = 
	  new cTempORTCNRData (NULL, ORTCData,NULL, NULL, 	
			       -1, (GetMask (ImmediatePrefix,depth)+1));
	ORTCData->child[RIGHT] = newORTC;
	newRIBNode = 
	  new cRIBNode (0, 0,next, NULL, NULL,RIB->pol);
	next->child[RIGHT] = newRIBNode;	
	(*newRIBNode).Data.SMALTAFlags.DistanceToParent = -1;
	(*newRIBNode).Data.SMALTAFlags.Original = 0;
	newleaves++;
      }
      flags[depth] = 1;
      if (ORTCData->child[LEFT] == NULL && ORTCData->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	ImmediatePrefix[depth] = 0;
	depth--;
	ORTCData = ORTCData->parent;
	next = next->parent;
	leaves ++;
      }
      else	{
	ORTCData = ORTCData->child[LEFT];
	next = next->child[LEFT];
	depth++;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
					
      }
      break;
		
    case 1:
      flags[depth] = 2;
      ORTCData = ORTCData->child[RIGHT];
      next = next->child[RIGHT];
      depth++;
      prefix = prefix | (1 << (32-depth));
      break;

    case 2:
      flags[depth] = 0;
      ORTCData = ORTCData->parent;
      next = next->parent;
      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      ImmediatePrefix[depth] = 0;
      depth--;
    }
  }
  std::cout << "Total leaves after pass 1 = " << leaves << "\n";
  //cout << "new leaves creates = " << newleaves << endl;
  return 0;
}
				


int cSMALTA::ORTCNRPass2 ()	{
  //std::cout << "ORTC Pass 2\n";
  for (unsigned int i = 0; i < NHtoIFaceMap.size(); i++)
    cout << NHtoIFaceMap[i] << endl;
  std::cout << NHtoIFaceMap.size() << "\n";
  cTempORTCNRData *ORTCData = TempORTCNRData;
  int depth = 0;
  intersections = 0;
  std::vector <int> flags (33,0);
  while (ORTCData != NULL)	{
    switch (flags[depth])	{
    case 0:
      if (ORTCData->child[LEFT] != NULL)	{
	flags[depth] = 1;
	ORTCData = ORTCData->child[LEFT];
	depth++;
      }
      else	{	// you are at leaf
	ORTCData->PotentialNHSet = CreateSet(ORTCData->IFace);
	ORTCData = ORTCData->parent;
	depth--;
      }
      break;
			
    case 1:
      flags[depth] = 2;
      ORTCData = ORTCData->child[RIGHT];
      depth++;
      break;	

    case 2:
      flags[depth] = 0;
      ORTCData->PotentialNHSet = 
	HashOp (ORTCData->child[LEFT]->PotentialNHSet,ORTCData->child[RIGHT]->PotentialNHSet);
      ORTCData = ORTCData->parent;
      depth--;
    }
  }
  std::cout << "Total intersections = " << intersections << "\n";			
  //std::cout << "Total Entries after ORTC = " << (leaves - intersections) << "\n";
  intersections = 0;	leaves = 0;
  return 0;
}

int cSMALTA::ORTCNRPass3 ()	{
  //std::cout << "ORTC Pass 3\n";
  cTempORTCNRData *temp,*ORTCData = TempORTCNRData;
  cRIBNode* next = RIB->root;
  cRIBNode* tempRIB;
  int m;
  int depth = 0; int tempif = 0; int DoNotWrite = 0;
  int deleted = 0;
  int generated = 0, artificial = 0;
  std::vector <int> flags (33,0);
  std::vector <int> ImmediatePrefixAT (33,0);
  int iter = 0;
  int g = 0;
  int r = 0;
  int encount = 0;
  while (ORTCData != NULL)	{
    switch (flags[depth])	{
    case 0:
      //iter++;
      //std::cout << "case 0 " << iter++ << "\n";
      //std::cout << "Current address " << ORTCData << " " << next << "\n";
      //cout << (*next).Data.Routes.BestNH << endl;
      if (ORTCData->parent != NULL && InSet (ORTCData->parent->IFace, ORTCData->PotentialNHSet))
	{	
	  //std::cout << "in set ";	
	  //std::cout << "Terminate current address if\n";
	  //std::cout << "RIBFlag " << (*next).Data.RIBFlags.INFIB << "\n";
	  ORTCData->IFace = ORTCData->parent->IFace;
	  // testing leaf, non orginal, and aggregated
	  //std::cout << "Terminate current address if2\n";
	  if (ORTCData->child[LEFT] == NULL && ORTCData->child[RIGHT] == NULL && ORTCData->ParentOffset != -1)	{
	    //std::cout << "artificial ";
	    artificial = 1;	
	    //cout << "Offset: " << ORTCData->ParentOffset << endl;
	    //cout << "Aggregated Prefix at: " << GetMask (ImmediatePrefixAT, depth) << endl;
	    //cout << "depth: " << depth << endl;
	    if (ORTCData->ParentOffset > GetMask (ImmediatePrefixAT, depth))	{
	      tempRIB = next;
	      m = GetMask (ImmediatePrefixAT,depth);
	      for (int i = 0; i < m; i++)	
		next = next->parent;
	      (*next).Data.SMALTAFlags.DistanceToParent = ORTCData->ParentOffset - m;
	      next = tempRIB;
	    }
	  }
	  //std::cout << "outside if\n";
	}
      else	{
	//std::cout << "Terminate current address else\n";
	tempif = ORTCData->IFace;
	ORTCData->IFace = choose (ORTCData->PotentialNHSet);
	if (next->Data.SMALTAFlags.Original == 0)	{		// not original entry
	  generated = 1;
	  g++;
	}
						
	else if (ORTCData->IFace != tempif)	{
	  if (InSet (tempif, ORTCData->PotentialNHSet))
	    ORTCData->IFace = tempif;
	  else	{
	    r++;
	    ORTCData->IFace = ORTCData->parent->IFace;
	    DoNotWrite = 1;
	  }
	}
					
	if (!DoNotWrite)	{
	  ImmediatePrefixAT[depth] = 1;
	  //cout << "Aggregated Prefix at depth: " << depth << endl;
	  (*next).Data.RIBFlags.INFIB = 1;
	  iter++;
	  //std::cout << "writing " << ORTCData->IFace << "\n";
	  if (generated == 1)	{
	    generated = 0;
	    (*next).Data.Routes.BestNH = NHtoIFaceMap[ORTCData->IFace];
	    (*next).Data.Routes.BestASLen = 0;
	    if ((*next).child[LEFT] == NULL && (*next).child[RIGHT] == NULL)
	      (*next).Data.SMALTAFlags.DistanceToParent = ORTCData->ParentOffset;
	  }
	  else
	    (*next).Data.SMALTAFlags.DistanceToParent = -1;
	}
				
	else	{
	  DoNotWrite = 0;
	  (*next).Data.RIBFlags.INFIB = 0;
	}
      }
      if (ORTCData->child[LEFT] == NULL && ORTCData->child[RIGHT] == NULL)	{
	//std::cout << "a leaf \n";
	encount++;
	flags[depth] = 0;
	ImmediatePrefixAT[depth] = 0;
	depth--;
	temp = ORTCData;
	tempRIB = next;
	ORTCData = ORTCData->parent;
	next = next->parent;
	if (artificial)	{		// take care of hanging pointers
	  artificial = 0;
	  //std::cout << "taking care of hanging pointer\n";
	  if (next->child[LEFT] == tempRIB)
	    next->child[LEFT] = NULL;
	  else 
	    next->child[RIGHT] = NULL;
	  delete tempRIB;
	  deleted++;
	}
	delete temp;
	// call ORTCData destructor
      }
      else {
	flags[depth] = 1;
					
	ORTCData = ORTCData->child[LEFT];
	next = next->child[LEFT];
	//std::cout << "not leaf\n";
	//std::cout << "traversing left " << ORTCData << " " << next << "\n";
	depth++;
      }
      break;
			
    case 1:
      //std::cout << "case 1\n";
      flags[depth] = 2;
      depth++;
      ORTCData = ORTCData->child[RIGHT];
      next = next->child[RIGHT];
      //std::cout << "traversing right " << ORTCData << " " << next << "\n";
      break;
		
    case 2:
      //std::cout << "case 2\n";
      flags [depth] = 0;
      ImmediatePrefixAT[depth] = 0;
      depth --;
      temp = ORTCData;
      ORTCData = ORTCData->parent;	
      next = next->parent;
      if (temp != TempORTCNRData)
	delete temp;
      // call ORTCData destructor
    }	// end switch
  }	// end while 
  TempORTCNRData->child[LEFT] = NULL;
  TempORTCNRData->child[RIGHT] = NULL;
  //cout << "leaves encountered = " << encount << endl;
  //cout << "leaves deleted = " << deleted << endl;
  return iter;
  //std::cout << "generated entries = " << g << "replaced entries = " << r << "\n";
}





int cSMALTA::InSet (int i, unsigned int* a)	{
  unsigned int l = a[i/32] >> (i%32);
  return (l%2);
	
}

int cSMALTA::choose (unsigned int* a)	{
  for (int i = 0; i < (int) NHtoIFaceMap.size();i++)	
    if (InSet (i, a))
      return i;
  return -1;
}

unsigned int* cSMALTA::CreateSet(int iface)	{
  int l = (NHtoIFaceMap.size() - 1)/32 + 1, j;
  unsigned int *a = new unsigned int [l];

  for (j = 0; j < l; j++)	{
    a[j] = 0;
    a[iface/32] = ((unsigned int) 1) << (iface % 32);
  }
  return a;
}

unsigned int* cSMALTA::HashOp (unsigned int* l,unsigned int* r)	{
  int w = (NHtoIFaceMap.size() - 1)/32 + 1, j;
  unsigned int* x = new unsigned int [w];
  int c = 1;
	
  for (j = 0; j < w; j++)	{
    x[j] = l[j] & r[j];
    if (x[j] != 0)	
      c = 0;
  }
  if (c == 0) {
    intersections++;
    return x;
  }
  else	{
    for (j = 0; j < w; j++)	
      x[j] = l[j] | r[j];
    return x;
  }
}

/************************************************************************************************************************************
SMALTA CODE
************************************************************************************************************************************/


int cSMALTA::SMALTAUpdate (unsigned int Prefix, int mask, unsigned int OldNH, int MessageType)	{
  if (MessageType == 1)	
    SMALTAAnnounce (Prefix, mask, OldNH, 1);
  else if (MessageType == 2)
    SMALTAAnnounce (Prefix, mask, OldNH, 2);
  else if (MessageType == 3)
    SMALTAAnnounce (Prefix, mask, OldNH, 3);
  else if (MessageType == 0)
    SMALTAWithdraw (Prefix, mask, OldNH);
  else	{
    cout << "Error ! Unknown message type exiting.. \n";
    exit(1);
  }
  //cout << "returning from SMALTAUpdate\n";
  return 0;
}

int cSMALTA::SMALTAWithdraw (unsigned int Prefix, int mask, unsigned int OldNH)	{
  cRIBNode *next = RIB->root, *P = NULL, *I = NULL, *N = NULL;
  int nbit = 32;
  int tmask = mask;
  int Pmask = 0, Imask = 0;
  int bit;
  unsigned int tempOldNH = OldNH;
  int Bool = 0;
  while (tmask > 0)	{			// preorder tree traversal to note down P and I

    if (next->Data.Routes.BestNH != NULL_NEXTHOP)	{
      if (next->Data.SMALTAFlags.Original == 1)	{
	Pmask = 32 - nbit;	
	P = next;
      }
      if (next->Data.RIBFlags.INFIB == 1)	{	
	I = next;	
	Imask = 32 - nbit;
      }
    }		
    nbit--;
				
    bit = (Prefix >> nbit) % 2;
    next = next->child[bit];
    tmask--;
		
  }
  N = next;
  //cout << "SMALTA Withdraw of" << Prefix << "/"<<mask<< " " << OldNH << " " << N->Data.RIBFlags.INFIB << "\n";
  if (P == NULL)	{
    cout << "P is NULL\n";
    exit(1);
  }
  if (I == NULL)	{
    cout << "I is NULL\n";
    exit(1);
  }
  if (N->Data.RIBFlags.INFIB == 0)
    OldNH = I->Data.Routes.BestNH;
  unsigned int NHtoCompare;
  if (N->Data.RIBFlags.INFIB == 1)	
    NHtoCompare = tempOldNH;
  else	{
    NHtoCompare = I->Data.Routes.BestNH;
    //Imask = mask;
    Bool = 1;
  }
  N->Data.SMALTAFlags.Original = 0;
  N->Data.RIBFlags.INFIB = 0;
  N->Data.SMALTAFlags.DistanceToParent = -1;
  //cout << "I is " << I->Data.Routes.BestNH<<"/"<<Imask << endl;
  //cout << "P is " << P->Data.Routes.BestNH <<"/"<<Pmask<< endl;
  /*
    if (next->Data.Routes.BestNH != NULL_NEXTHOP)	{
    if (next->Data.SMALTAFlags.Original == 1)	{
    Pmask = 32 - nbit;	
    P = next;
    }
    if (next->Data.RIBFlags.INFIB == 1)	{	
    I = next;	
    Imask = 32 - nbit;
    }
    }
  */	
  if (P->Data.Routes.BestNH != I->Data.Routes.BestNH)	{
    N->Data.RIBFlags.INFIB = 1;
    N->Data.Routes.BestNH = P->Data.Routes.BestNH;
    N->Data.SMALTAFlags.DistanceToParent = mask - Pmask;
    NodesChanged++;
  }
  else if (Pmask < Imask)	{
    if (Bool == 1)
      I->Data.SMALTAFlags.DistanceToParent = Imask - Pmask;
    //cout << "Pmask < Imask\n";
    if (N->child[LEFT] == NULL && N->child[RIGHT] == NULL)	{		// A Leaf
      //cout << "deleting leaf\n";
      cRIBNode *temp = N->parent;
      if (temp->child[LEFT] == N)
	temp->child[LEFT] = NULL;
      else if (temp->child[RIGHT] == N)
	temp->child[RIGHT] = NULL;
      else	{
	cout << "Addressing error \n";
	exit(1);
      }
      delete N;
			
      NodesChanged++;
      return 0;
    }
  }
  // At this stage we have P and I. Executing the SMALTA Algorithm
  if (P->Data.Routes.BestNH != NHtoCompare)	{
    //cout << "N is " << N->Data.Routes.BestNH << endl;
		
    next = N;
    // deaggregate the immediate descendents
    //cout << "deaggregate the immediate descendents\n";
    cRIBNode *terminatingNode = N->parent;
    vector <int> flags (33,0);
    int depth = mask;
    int TraverseDown = 0;
    while (next != terminatingNode)	{
      switch (flags[depth])	{
      case 0:
					
	if (next != N && next->Data.Routes.BestNH != NULL_NEXTHOP)	{
	  //cout << "considering " << next->Data.Routes.BestNH << endl;
	  if (next->Data.Routes.BestNH == N->Data.Routes.BestNH)	{	// d[i] = d[n]
	    //cout << "aggregating\n";
	    if (next->Data.SMALTAFlags.Original == 0)	{		// generated entry
	      next->Data.Routes.BestNH = NULL_NEXTHOP;
	      next->Data.RIBFlags.INFIB = 0;
	      next->Data.SMALTAFlags.DistanceToParent = -1;	
	      NodesChanged++;
	    }
	    else {
	      if (next->Data.RIBFlags.INFIB == 1)	{
		next->Data.RIBFlags.INFIB = 0;
		NodesChanged++;
	      }
	      else
		TraverseDown = 1;
	    }
	  }
	  else	{
	    if (next->Data.RIBFlags.INFIB == 0 && next->Data.Routes.BestNH == OldNH)	{
	      NodesChanged++;
	      next->Data.RIBFlags.INFIB = 1;
	      //cout << "restoring :" << next->Data.Routes.BestNH <<"/"
	      //<< depth << endl;
	      if (next->Data.SMALTAFlags.Original == 0)	{
		cout << "error ! A generated entry was non FIB !" << endl;
		exit(1);
	      }
	    }
	    else if (next->Data.RIBFlags.INFIB == 0 && next->Data.Routes.BestNH != OldNH)
	      TraverseDown = 1;
	  }
	  if (!TraverseDown)	{
	    flags[depth] = 2;
	    break;
	  }
	  TraverseDown = 0;
	}
	flags[depth] = 1;
	if (next->child[LEFT] == NULL)	{
	  flags[depth] = 2;
	  if (next->child[RIGHT] == NULL)	{
	    flags[depth] = 0;	
	    next = next->parent;
	    depth--;
	  }
	  else	{
	    next = next->child[RIGHT];
	    depth++;
	  }
	}
	else	{	
	  next = next->child[LEFT];
	  depth++;
	}
	break;
				
      case 1:
	flags [depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;	
	  next = next->parent;
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	}
	break;
				
      case 2:
	flags[depth] = 0;
	next = next->parent;
	depth--;
      }
    }
  }
	
  //else	
  //cout << "PMask !< Imask\n";
  OldNH = tempOldNH;
  next = N;
  int depth = mask;
  cRIBNode* terminatingNode = N->parent;
  vector <int> flags (33,0);
  //cout << "Taking care of deaggregates\n";
  while (next != terminatingNode)	{
    switch (flags[depth])	{
    case 0:
      if (next != N && next->Data.Routes.BestNH != NULL_NEXTHOP)	{
	if (next->Data.RIBFlags.INFIB == 1 && next->Data.SMALTAFlags.DistanceToParent == (depth - mask))	{	// a deaggregate of N
	  if (next->Data.SMALTAFlags.Original == 1)	{
	    cout << "Error ! An original considered as a deaggregate" << endl;
	    exit(1);
	  }
	  next->Data.SMALTAFlags.DistanceToParent = depth - Pmask;
	  next->Data.Routes.BestNH = P->Data.Routes.BestNH;
	  //cout << "Deaggregate encountered: " << next->Data.Routes.BestNH <<"/"<<depth<< 
	  //" "<<next->Data.RIBFlags.INFIB<<" "<<next->Data.SMALTAFlags.Original<<" "
	  //<<next->Data.SMALTAFlags.DistanceToParent<< endl;
	  NodesChanged++;
	  if (next->Data.Routes.BestNH != P->Data.Routes.BestNH)	{
	    cout << "Error ! Routing table correcness compromised by deaggregates\n";
	    exit(1);
	  }
	  // take care of immediate descendents of the deaggregates
	  //cout << "take care of immediate descendents of the deaggregates\n";
	  {
	    int rdepth = depth;
	    cRIBNode* rN = next;
	    cRIBNode* rnext = next;
	    cRIBNode* rterminate = rN->parent;
	    vector <int> rflags (33,0);
	    while (rnext != rterminate)	{
	      switch (rflags[rdepth])	{
	      case 0:
		if (rnext!=rN && rnext->Data.Routes.BestNH == rN->Data.Routes.BestNH){
		  if (rnext->Data.RIBFlags.INFIB == 1)	{
		    rnext->Data.RIBFlags.INFIB = 0;
		    NodesChanged++;	
		    rflags[rdepth] = 2;
		    break;
											
		  }
										
                                                              				
		}
		else 	{
		  if (rnext!=rN && rnext->Data.Routes.BestNH != rN->Data.Routes.BestNH){
		    if (rnext->Data.RIBFlags.INFIB == 0 && 
			rnext->Data.Routes.BestNH == OldNH)	{
		      rnext->Data.RIBFlags.INFIB = 1;
		      NodesChanged++;	
		      rflags[rdepth] = 2;
		      break;
											
		    }
		    if (rnext->Data.RIBFlags.INFIB == 1 && 
			rnext->Data.Routes.BestNH != OldNH)	{
		      rflags[rdepth] = 2;
		      break;
		    }
										
		  }
		}


		rflags[rdepth] = 1;
		if (rnext->child[LEFT] == NULL)	{
		  rflags[rdepth] = 2;
		  if (rnext->child[RIGHT] == NULL)	{
		    rflags[rdepth] = 0;
		    rnext = rnext->parent;	
		    rdepth--;
		  }
		  else	{
		    rnext = rnext->child[RIGHT];
		    rdepth++;
		  }
		}
		else	{
		  rnext = rnext->child[LEFT];
		  rdepth++;
		}
		break;
						
	      case 1:
		rflags[rdepth] = 2;	
		if (rnext->child[RIGHT] == NULL)	{
		  rflags[rdepth] = 0;
		  rnext = rnext->parent;	
		  rdepth--;
		}
		else	{
		  rnext = rnext->child[RIGHT];
		  rdepth++;
		}
		break;
		
	      case 2:
		rflags[rdepth] = 0;
		rnext = rnext->parent;
		rdepth--;
	      }
	    }
	  }
			
	}
      }
      flags[depth] = 1;
      if (next->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  next = next->parent;	
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	}
      }
      else	{
	next = next->child[LEFT];
	depth++;
      }
      break;
			
    case 1:
      flags[depth] = 2;
      if (next->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = next->parent;
	depth--;
      }
      else	{
	next = next->child[RIGHT];
	depth++;
      }
      break;
			
    case 2:
      flags[depth] = 0;
      next = next->parent;
      depth--;
    }
  }
  return 0;
}

int cSMALTA::SMALTAAnnounce (unsigned int Prefix, int mask, unsigned int OldNH, int subMessage)	{
	
  cRIBNode *next = RIB->root;
  //cout << "SMALTA Announce2\n";
  cRIBNode *P = NULL, *I = NULL, *N = NULL, *TerminatingNode = NULL;
  int nbit = 32;
  int tmask = mask;
  int Pmask = 0, Imask = 0;
  int bit;
  int NinOT;
  unsigned int tempOldNH = OldNH;
  //cout << mask << " " << Prefix << "\n";
  while (tmask > 0)	{			// preorder tree traversal to note down P and I
    //cout << "A " << next->Data.Routes.BestNH << " "<< next->Data.SMALTAFlags.Original << " " << next->Data.RIBFlags.INFIB << "\n";
    if (next->Data.Routes.BestNH != NULL_NEXTHOP)	{
      if (next->Data.SMALTAFlags.Original == 1)	{
	Pmask = 32 - nbit;	
	P = next;
      }
      if (next->Data.RIBFlags.INFIB == 1)	{	
	I = next;	
	Imask = 32 - nbit;
      }
    }		
    nbit--;
				
    bit = (Prefix >> nbit) % 2;
    next = next->child[bit];
    tmask--;
		
  }

	
  N = next;
  //cout << "SMALTA Announce of" << Prefix << "/"<<mask<< " " << N->Data.Routes.BestNH << "\n";
  TerminatingNode = N->parent;
  N->Data.SMALTAFlags.Original = 1;
  // At this stage we have P and I. Executing the SMALTA Algorithm
  //if (subMessage == 2 || subMessage == 3)
  if (subMessage == 2)	
    OldNH = I->Data.Routes.BestNH;
  if (subMessage == 1)	{
    if (N->Data.RIBFlags.INFIB == 0)
      OldNH = I->Data.Routes.BestNH;
  }
  if (subMessage == 1)	
    NinOT = 1;
  else 
    NinOT = 0;
  if (NinOT == 1)	{
    Pmask = 32 - nbit;	
    P = next;
  }
			
  //cout << "I is " << I->Data.Routes.BestNH <<"/"<<Imask<< endl;
  //cout << "P is " << P->Data.Routes.BestNH <<"/"<<Pmask<< endl;
	
  //cout << "Gotten P and I\n";
  if (I == NULL)	{
    cout << "error ! I is NULL\n";
    exit(1);
  }
  if (I->Data.Routes.BestNH == N->Data.Routes.BestNH)	{
    //cout << "I and N match\n";
    NodesChanged++;
    N->Data.RIBFlags.INFIB = 0;
  }
  if (subMessage == 3 || subMessage == 1 || I->Data.Routes.BestNH != N->Data.Routes.BestNH)	{
    //cout << "I and N do not match, mask " << mask << "\n";
    next = N;
    if (!(I->Data.Routes.BestNH == N->Data.Routes.BestNH))
      N->Data.RIBFlags.INFIB = 1;
    vector <int> flags (33,0);
    int depth = mask;
    int TraverseDown = 0;
    while (next != TerminatingNode)	{
      switch (flags[depth])	{	
      case 0:
	//cout << next << endl;
	if (next != N && next->Data.Routes.BestNH != NULL_NEXTHOP)	{
	  //cout << "Considering " << next->Data.Routes.BestNH <<"/"<<depth<<
	  //" " <<next->Data.RIBFlags.INFIB<<" "<<next->Data.SMALTAFlags.Original<<" "
	  //<<next->Data.SMALTAFlags.DistanceToParent<< endl;
	  if (next->Data.Routes.BestNH == N->Data.Routes.BestNH)	{	// d[i] = d[n]
	    if (next->Data.SMALTAFlags.Original == 0)	{		// generated entry
	      next->Data.Routes.BestNH = NULL_NEXTHOP;
	      NodesChanged++;
	      next->Data.RIBFlags.INFIB = 0;
	      next->Data.SMALTAFlags.DistanceToParent = -1;
	    }
	    else	{
	      if (next->Data.RIBFlags.INFIB == 1)	{
		NodesChanged++;
		next->Data.RIBFlags.INFIB = 0;
	      }
	      else	
		TraverseDown = 1;
	    }
	  }
	  else	{
	    if (next->Data.RIBFlags.INFIB == 0 && next->Data.Routes.BestNH == OldNH)	{
	      next->Data.RIBFlags.INFIB = 1;
	      //cout << "restoring " << next->Data.Routes.BestNH<<"/"<<depth << 
	      //" " <<next->Data.RIBFlags.INFIB<<" "<<next->Data.SMALTAFlags.Original<<" "
	      //<<next->Data.SMALTAFlags.DistanceToParent<< endl;
	      NodesChanged++;
	    }
	    else if (next->Data.RIBFlags.INFIB == 0 && next->Data.Routes.BestNH != OldNH)
	      TraverseDown = 1;
	  }
	  if (!TraverseDown)	{
	    flags[depth] = 2;
	    break;
	  }
	  TraverseDown = 0;
	}
	flags[depth] = 1;
	if (next->child[LEFT] == NULL)	{
	  flags[depth] = 2;
	  if (next->child[RIGHT] == NULL)	{
	    flags[depth] = 0;	
	    next = next->parent;
	    depth--;
	  }
	  else	{
	    next = next->child[RIGHT];
	    depth++;
	  }
	}
	else	{	
	  next = next->child[LEFT];
	  depth++;
	}
	break;
				
      case 1:
	flags [depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;	
	  next = next->parent;
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	}
	break;
				
      case 2:
	flags[depth] = 0;
	next = next->parent;
	depth--;
      }	// end switch
    }	// end while
  }	// end else
	//cout << "Taking care of deaaggregates\n";
  cRIBNode *T;
  OldNH = tempOldNH;
  if (subMessage == 2 || subMessage == 3)
    OldNH = P->Data.Routes.BestNH;
  if (NinOT == 1)
    T = N;
  else
    T = P;
  int Tmask;
  if (T == N)	
    Tmask = mask;
  else if (T == P)	
    Tmask = Pmask;
  cRIBNode* terminatingNode = N->parent;
  vector <int> flags (33,0);
  int depth = mask;
  next = N;
  while (next != terminatingNode)	{
    switch (flags[depth])	{
    case 0:
      //cout << next << " " << N << " " << next->child[LEFT] << " " << next->child[RIGHT] << endl;
      if (next != N && next->Data.Routes.BestNH != NULL_NEXTHOP)	{
	if (next->Data.RIBFlags.INFIB == 1 && next->Data.SMALTAFlags.DistanceToParent == (depth - Tmask))	{	// a deaggregate of T
	  if (next->Data.SMALTAFlags.Original == 1)	{
	    cout << "Error ! An original entry considered a deaggregate" << endl;
	    exit(1);
	  }
	  //cout << "Deaggregate encountered: " << next->Data.Routes.BestNH<<"/"<<depth<< 
	  //" "<<next->Data.RIBFlags.INFIB<<" "<<next->Data.SMALTAFlags.Original<<" "
	  //<<next->Data.SMALTAFlags.DistanceToParent<< endl;
	  next->Data.SMALTAFlags.DistanceToParent = depth - mask;
	  next->Data.Routes.BestNH = N->Data.Routes.BestNH;
						
	  NodesChanged++;
	  if (next->Data.Routes.BestNH != N->Data.Routes.BestNH)	{
	    cout << "Error ! Routing table correcness compromised by deaggregates\n";
	    exit(1);
	  }
	  // take care of immediate descendents of the deaggregates
	  if (!(next->child[LEFT] == NULL && next->child[RIGHT] == NULL))	{
	    //cout << "take care of immediate descendents of the deaggregates\n";
	    int rdepth = depth;
	    cRIBNode* rN = next;
	    cRIBNode* rnext = next;
	    cRIBNode* rterminate = rN->parent;
	    vector <int> rflags (33,0);
	    int TraverseDown = 0;
	    while (rnext != rterminate)	{
	      switch (rflags[rdepth])	{
	      case 0:
		//cout << rdepth << "\n";
		if (rnext!=rN && rnext->Data.Routes.BestNH != NULL_NEXTHOP)	{
		  if( rnext->Data.Routes.BestNH == rN->Data.Routes.BestNH) {
		    if (rnext->Data.RIBFlags.INFIB == 1)	{
		      if (rnext->Data.SMALTAFlags.Original == 0) {
			rnext->Data.Routes.BestNH = NULL_NEXTHOP;
			NodesChanged++;
			rnext->Data.RIBFlags.INFIB = 0;
			rnext->Data.SMALTAFlags.DistanceToParent = -1;
		      }
		      else {
			rnext->Data.RIBFlags.INFIB = 0;
			NodesChanged++;
		      }
		    }
		    else
		      TraverseDown = 1;
											
		  }
		  else	{
		    if (rnext->Data.RIBFlags.INFIB == 0 &&
			rnext->Data.Routes.BestNH == OldNH)	{
		      rnext->Data.RIBFlags.INFIB = 1;
		      NodesChanged++;
		    }
		    else if (rnext->Data.RIBFlags.INFIB == 0 &&
			     rnext->Data.Routes.BestNH != OldNH)		
		      TraverseDown = 1;
		  }
		  if (!TraverseDown)	{
		    rflags[rdepth] = 2;
		    break;
		  }
		  TraverseDown = 0;
		}
		rflags[rdepth] = 1;
		if (rnext->child[LEFT] == NULL)	{
		  rflags[rdepth] = 2;
		  if (rnext->child[RIGHT] == NULL)	{
		    rflags [rdepth] = 0;
		    rnext = rnext->parent;	
		    rdepth--;
		  }
		  else	{
		    rnext = rnext->child[RIGHT];
		    rdepth++;
		  }
		}
		else	{
		  rnext = rnext->child[LEFT];
		  rdepth++;
		}
		break;
						
	      case 1:
									
		rflags[rdepth] = 2;	
		if (rnext->child[RIGHT] == NULL)	{
		  rflags[rdepth] = 0;
		  rnext = rnext->parent;	
		  rdepth--;
		}
		else	{
		  rnext = rnext->child[RIGHT];
		  rdepth++;
		}
		break;
		
	      case 2:
		rflags[rdepth] = 0;
		rnext = rnext->parent;
		rdepth--;
	      }
	    }
	    //cout << "mmediate descendents of deaggregates taken care of\n";
	  }
	}
      }
      flags[depth] = 1;
      if (next->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  next = next->parent;	
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	}
      }
      else	{
	next = next->child[LEFT];
	depth++;
      }
      break;
			
    case 1:
      flags[depth] = 2;
      if (next->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = next->parent;
	depth--;
      }
      else	{
	next = next->child[RIGHT];
	depth++;
      }
      break;
			
    case 2:
      flags[depth] = 0;
      next = next->parent;
      depth--;
    }
  }
	
  //cout << "Announce Successfull\n";
  return 0;
}	


/***************************************************************************************************************************
Aggregation Algorithms according to the IETF draft "draft-zhang-fibaggregation-02"
***************************************************************************************************************************/

int cZhangAgg::Apply()	{
  switch (AggLevel)	{
  case 1:
    ApplyLevel1();
    break;
  case 2:
    ApplyLevel2();
    break;
  }
  return 0;
}


void cZhangAgg::ApplyLevel1()	{
  cRIBNode* next = RIB->root;
  vector <unsigned int> ImmParentNH (33,NULL_NEXTHOP);
  vector <int> flags (33,0);
  int depth = 0;

  while (next != NULL)	{
    switch (flags[depth])	{
    case 0:
      if (next->Data.Routes.BestNH != NULL_NEXTHOP)	{
	next->Data.SMALTAFlags.Original = 1;
	next->Data.RIBFlags.INFIB = 1;
	if (next->Data.Routes.BestNH == GetImmAncestorNH(ImmParentNH, depth))
	  next->Data.RIBFlags.INFIB = 0;
	else
	  ImmParentNH[depth] = next->Data.Routes.BestNH;
      }
      flags[depth] = 1;
      if (next->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags [depth] = 0;
	  next = next->parent;
	  ImmParentNH[depth] = NULL_NEXTHOP;
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	}
      }
      else	{
	next = next->child[LEFT];
	depth++;
      }
      break;
		
    case 1:
      flags[depth] = 2;
      if (next->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = next->parent;
	ImmParentNH[depth] = NULL_NEXTHOP;
	depth--;
      }
      else	{
	next = next->child[RIGHT];
	depth++;
      }
      break;
				
    case 2:
      flags[depth] =  0;
      next = next->parent;
      ImmParentNH[depth] = NULL_NEXTHOP;
      depth--;

    }
  }
}

void cZhangAgg::ApplyLevel2()	{
  cRIBNode* next = RIB->root;
  vector <int> flags (33,0);
  int depth = 0;
  ApplyLevel1();
  while (next!= NULL)	{
    switch (flags[depth])	{
    case 0:	
      flags[depth] = 1;
      /*
	if (next->Data.Routes.BestNH != NULL_NEXTHOP)	{
	next->Data.SMALTAFlags.Original = 1;
	next->Data.RIBFlags.INFIB = 1;
	}
	else	{
	next->Data.SMALTAFlags.Original = 0;
	next->Data.RIBFlags.INFIB = 0;
	}
      */
      if (next->child[LEFT] != NULL)	{	
	next = next->child[LEFT];
	depth++;
      }
				
      break;
		
    case 1:
      flags[depth] = 2;
      if (next->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = next->parent;
	depth--;
      }
      else	{
	next = next->child[RIGHT];
	depth++;
      }
      break;
				
    case 2:
      if (next->child[LEFT]!=NULL && next->child[RIGHT]!=NULL && next->Data.Routes.BestNH == NULL_NEXTHOP)	{
	if ((next->child[LEFT]->Data.Routes.BestNH == next->child[RIGHT]->Data.Routes.BestNH) &&
	    (next->child[LEFT]->Data.Routes.BestNH != NULL_NEXTHOP))	{
	  next->Data.Routes.BestNH = next->child[LEFT]->Data.Routes.BestNH;
	  next->Data.RIBFlags.INFIB = 1;
	  next->Data.SMALTAFlags.Original = 0;
	  if (next->child[LEFT]->Data.SMALTAFlags.Original == 0)	{
	    next->child[LEFT]->Data.Routes.BestNH = NULL_NEXTHOP;
							
	  }
	  if (next->child[RIGHT]->Data.SMALTAFlags.Original == 0)	{
	    next->child[RIGHT]->Data.Routes.BestNH = NULL_NEXTHOP;
							
	  }
	  next->child[LEFT]->Data.RIBFlags.INFIB = 0;
	  next->child[RIGHT]->Data.RIBFlags.INFIB = 0;
	}
      }
      flags[depth] =  0;
      next = next->parent;
      depth--;

    }
  }
					
}

unsigned int cZhangAgg::GetImmAncestorNH (vector <unsigned int> &ImmAnc, int d)	{
  for (int i = d-1; i >= 0; i--)	{
    if (ImmAnc[i] != NULL_NEXTHOP)	
      return ImmAnc[i];
  }
  return NULL_NEXTHOP;
}	

/**********************************************************************************************************************
Update Algorithms for zhang Aggregation
**********************************************************************************************************************/

void cZhangAgg::Update (unsigned int Prefix, int Mask, int Message)	{
  cRIBNode *n = RIB->root;
  int nbit = 32;
  int mask = Mask;
  cRIBNode* P = NULL;
  int bit;
  int Pmask = -1;
  //cout << "Zhang update " << mask << " " << w << " \n";
  while (mask > 0)	{
		
    nbit--;
		
    if (n->Data.Routes.BestNH != NULL_NEXTHOP)	{	
      P = n;
      Pmask = Mask - mask;
    }
		
    bit = (Prefix >> nbit) % 2;
    if (n == NULL)	{
      cout << "n is NULL\n";
      exit(1);
    }
    n = n->child[bit];
    mask--;
  }
  n->Data.RIBFlags.INFIB = 1;

  switch (AggLevel)	{
  case 1:
    if (Message == 1 || Message == 2)	{
      if (P != NULL)	{
	if (P->Data.Routes.BestNH == n->Data.Routes.BestNH)	{
	  if (n->Data.RIBFlags.INFIB == 1)	{	
	    n->Data.RIBFlags.INFIB = 0;
	    NodesLevel1++;
	  }
	}
	else	{	
	  if (n->Data.RIBFlags.INFIB == 0)	{
	    n->Data.RIBFlags.INFIB = 1;
	    NodesLevel1++;
	  }
	}
      }
      else 	{
	if (n->Data.RIBFlags.INFIB == 0)	{
	  n->Data.RIBFlags.INFIB = 1;
	  NodesLevel1++;
	}
      }
      n->Data.SMALTAFlags.Original = 1;
      reAggregate (n, Mask);
    }
    else if (Message == 0)	{
      n->Data.RIBFlags.INFIB = 0;
      NodesLevel1++;
      deAggregate (n, P, Mask);
    }
    else if (Message == 3)	{
      cout << "Error ! A generated entry is not supposed to exist for level 1\n";
      exit(1);
    }
    break;

  case 2:
    if (Message == 1 || Message == 2 || Message == 3)	{
      if (P!=NULL)	{
	if (P->Data.Routes.BestNH == n->Data.Routes.BestNH)	{	
	  if (n->Data.RIBFlags.INFIB == 1)	{
	    n->Data.RIBFlags.INFIB = 0;
	    NodesLevel2++;
	  }
	}
	else	{
	  if (n->Data.RIBFlags.INFIB == 0)	{
	    n->Data.RIBFlags.INFIB = 1;
	    NodesLevel2++;
	  }
	}
      }
      else	{
	if (n->Data.RIBFlags.INFIB == 0)	{
	  n->Data.RIBFlags.INFIB = 1;
	  NodesLevel2++;
	}
      }
      n->Data.SMALTAFlags.Original = 1;
      reAggregate (n, Mask);
    }
    else if (Message == 0)	{
      n->Data.RIBFlags.INFIB = 0;
      NodesLevel2++;
      //cout << "inside mesage 0\n";
      deAggregate (n, P, Mask);
      if (P != NULL)	{
	if (P->Data.SMALTAFlags.Original == 0)	{
	  n = RIB->root;
	  nbit = 32;
	  cRIBNode *A = NULL;
	  int Amask = 0;
	  int k = 0;
	  while (n!= P)	{
	    nbit--;
	    if (n->Data.Routes.BestNH != NULL_NEXTHOP)	{	
	      A = n;
	      Amask = k;
	    }
	    bit = (Prefix >> nbit) % 2;
	    n = n->child[bit];
	    k++;
	  }
	  deAggregate (P, A, Pmask);
	  P->Data.Routes.BestNH = NULL_NEXTHOP;
	  P->Data.RIBFlags.INFIB = 0;
	  P->Data.SMALTAFlags.Original = 0;
						
	}
      }
    }
  }
}
					
							

void cZhangAgg::reAggregate (cRIBNode* node, int mask)	{
  cRIBNode *t = node->parent;
  cRIBNode *n = node;
  vector <int> flags (33,0);
  int depth = mask;
  while (n != t)	{
    switch (flags[depth])	{
    case 0:
      if (n != node && n->Data.Routes.BestNH != NULL_NEXTHOP)	{
	if (n->Data.Routes.BestNH == node->Data.Routes.BestNH)	{
	  if (n->Data.RIBFlags.INFIB == 1)	{
	    if (AggLevel == 1)
	      NodesLevel1++; 
	    else if (AggLevel == 2)
	      NodesLevel2++;
	    n->Data.RIBFlags.INFIB = 0;
	  }
	}
	else 	{
	  if (n->Data.RIBFlags.INFIB == 0)	{
	    n->Data.RIBFlags.INFIB = 1;
	    if (AggLevel == 1)
	      NodesLevel1++; 
	    else if (AggLevel == 2)
	      NodesLevel2++;
	  }
	}
	flags[depth] = 2;
	break;
      }
      flags[depth] = 1;
      if (n->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (n->child[RIGHT] == NULL)	{
	  flags [depth] = 0;
	  n = n->parent;
	  depth--;
	}
	else	{
	  n = n->child[RIGHT];
	  depth++;
	}
      }
      else	{
	n = n->child[LEFT];
	depth++;
      }
      break;

    case 1:
      flags[depth] = 2;
      if (n->child[RIGHT] == NULL)	{
	flags [depth] = 0;
	n = n->parent;
	depth--;
      }
      else	{
	n = n->child[RIGHT];
	depth++;
      }
      break;
		
    case 2:
      flags [depth] = 0;
      n = n->parent;
      depth--;
    }
  }
} 

void cZhangAgg::deAggregate (cRIBNode *N, cRIBNode *P, int Mask)	{
  cRIBNode *T = N->parent;
  cRIBNode *n = N;
  vector <int> flags (33,0);
  int depth = Mask;
  while (n != T)	{
    switch (flags[depth])	{
    case 0:
      if (n != N && n->Data.Routes.BestNH != NULL_NEXTHOP)	{
	if (P !=NULL)	{
	  if (n->Data.Routes.BestNH == P->Data.Routes.BestNH)	{
	    if (n->Data.RIBFlags.INFIB == 1)	{
	      n->Data.RIBFlags.INFIB = 0;
	      if (AggLevel == 1)
		NodesLevel1++; 
	      else if (AggLevel == 2)
		NodesLevel2++;
	    }
	  }
	  else	{
	    if ( n->Data.RIBFlags.INFIB == 0)	{
	      n->Data.RIBFlags.INFIB = 1;
	      if (AggLevel == 1)
		NodesLevel1++;
	      else if (AggLevel == 2) 
		NodesLevel2++;
	    }
	  }
						
	}
	else	{
	  n->Data.RIBFlags.INFIB = 1;
	}
	flags[depth] = 2;
	break;
      }
      flags[depth] = 1;
      if (n->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (n->child[RIGHT] == NULL)	{
	  flags [depth] = 0;
	  n = n->parent;
	  depth--;
	}
	else	{
	  n = n->child[RIGHT];
	  depth++;
	}
      }
      else	{
	n = n->child[LEFT];
	depth++;
      }
      break;
		
    case 1:
      flags[depth] = 2;
      if (n->child[RIGHT] == NULL)	{
	flags [depth] = 0;
	n = n->parent;
	depth--;
      }
      else	{
	n = n->child[RIGHT];
	depth++;
      }
      break;
		
    case 2:
      flags [depth] = 0;
      n = n->parent;
      depth--;
    }
  }
}

/*****************************************************************************************************************

ORTC Based Algorithms for one-shot and update

*****************************************************************************************************************/

int cORTCUpdate::ApplyORTC()	{
  //cout << "applying ortc\n";
  int num_entries = 0;
  ReadRIB();
  cout << "Now starting Pass 1..." << endl;
  ORTCPass1();
  cout << "Now starting Pass 2..." << endl;
  ORTCPass2();
  cout << "Now starting Pass 3..." << endl;
  num_entries = ORTCPass3();
  return num_entries;
}

int cORTCUpdate::GetMask (std::vector<unsigned int>& ImmPref, int d)	{
  for (int i = d; i >= 0; i--)	
    if (ImmPref[i] == 1)	
      return (d - i);
  return -1;
}

int cORTCUpdate::ConvertNHtoIFace (unsigned int NH)	{
  for (int i = 0; i <(int) NHToIFaceMap.size(); i++)	{
    if (NH == NHToIFaceMap[i])	{
      return i;
    }
  }
  return -2;
}

int cORTCUpdate::InSet (int i, unsigned int* a)	{
  unsigned int l = a[i/32] >> (i%32);
  return (l%2);
	
}

int cORTCUpdate::choose (unsigned int* a)	{
  for (int i = 0; i < (int)NHToIFaceMap.size();i++)	
    if (InSet (i, a))
      return i;
  return -1;
}

unsigned int* cORTCUpdate::CreateSet(int iface)	{
  int l = (NHToIFaceMap.size() - 1)/32 + 1, j;
  unsigned int *a = new unsigned int [l];

  for (j = 0; j < l; j++)	{
    a[j] = 0;
    a[iface/32] = ((unsigned int) 1) << (iface % 32);
  }
  return a;
}

unsigned int* cORTCUpdate::HashOp (unsigned int* l,unsigned int* r)	{
  int w = (NHToIFaceMap.size() - 1)/32 + 1, j;
  unsigned int* x = new unsigned int [w];
  int c = 1;
	
  for (j = 0; j < w; j++)	{
    x[j] = l[j] & r[j];
    if (x[j] != 0)	
      c = 0;
  }
  if (c == 0) {
    inter++;
    return x;
  }
  else	{
    for (j = 0; j < w; j++)	
      x[j] = l[j] | r[j];
    return x;
  }
}

int cORTCUpdate::ReadRIB()	{
  cRIBdualNHNode *Rib = RIBdualNH->root;
  cTempORTCNRData *tData;
  cRIBdualNHNode *trib = NULL;
  cTempORTCNRData *temp = NULL;
  vector <int> flags (33,0);
  int depth = 0;
  int match = 0;
  int iter = 0;
  int numNH = 0;       // debug thingy
  unsigned int prefix = 0;
  // introduce drop at the root
  if (Rib->BestOrigNH == NULL_NEXTHOP)	{
    Rib->BestOrigNH = DROP;
    Rib->BestASLen = 0;
  }

  // initializing statistics
  nodes_changed_ortc = 0;  // Application of ortc sets nodes changed to 0
  nodes_changed_smalta = 0;	// smalta updated is applied AFTER ortc, hence initialize the nodes changed to 0
  effective_updates = 0;
  for (int j = 0; j < 33; j++)	{
    effective_updates_by_prefix[j] = 0;
    nodes_changed_by_prefix[j] = 0;
  }

  TempData = new cTempORTCNRData (NULL, NULL, NULL, NULL, 0, -1);
  tData = TempData;
  // cout << TempData << endl;
  while (Rib != NULL)	{
    switch (flags[depth])	{
    case 0:
      //if (prefix == 134217728)
      //cout << "depth = " << depth << " Rib left = " << Rib->child[LEFT] << "tdata left = "<< tData->child[LEFT] << endl;
      //cout << prefix << "/" << depth << " " << tData << " " << Rib << endl; 
      if (Rib->BestOrigNH != NULL_NEXTHOP)	{
	for (int i = 0; i < (int) NHToIFaceMap.size(); i++)	{
	  if (Rib->BestOrigNH == NHToIFaceMap[i])
	    match = 1;
	}
	if (!match)
	  {
	    NHToIFaceMap.push_back (Rib->BestOrigNH);
	    numNH++;
	  }
	tData->IFace = ConvertNHtoIFace (Rib->BestOrigNH);
	if ((*tData).IFace == -2)	{
	  cout << "Error in converting NH " << Rib->BestOrigNH <<  "to Interface. Exiting... \n";	
	  exit(1);
	}
	match = 0;
      }
      /*      if (Rib->child[LEFT] == NULL && Rib->child[RIGHT] == NULL)
	if (Rib->BestOrigNH == NULL_NEXTHOP)	{
	  iter--;
	  trib = Rib;
	  temp = tData;
	  tData = tData->parent;
	  Rib = Rib->parent;
	  depth--;
	  if (Rib->child[LEFT] == trib)	{
	    Rib->child[LEFT] = NULL;
	    tData->child[LEFT] = NULL;
	    delete trib;
	    delete temp;
	  }
	  else if (Rib->child[RIGHT] == trib)	{
	    Rib->child[RIGHT] = NULL;
	    tData->child[RIGHT] = NULL;
	    delete trib;
	    delete temp;
	  }
	  break;
	  }*/
							
      flags[depth] = 1;
      if (Rib->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (Rib->child[RIGHT] == NULL)	{
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  flags[depth] = 0;
	  Rib = Rib->parent;
	  tData = tData->parent;
	  depth--;
						
	}
	else	{
	  Rib = Rib->child[RIGHT];
	  cTempORTCNRData *t = new cTempORTCNRData (NULL,tData,NULL,NULL,-1,-1);
	  iter++;
	  tData->child[RIGHT] = t;
	  tData = tData->child[RIGHT];
	  depth++;
	  prefix = prefix | (1 << (32-depth));
	  //cout << "created node at depth " << depth << endl;
						
						
	}
      }
      else	{
	Rib = Rib->child[LEFT];
	cTempORTCNRData *t = new cTempORTCNRData (NULL,tData,NULL,NULL,-1,-1);
	iter++;
	tData->child[LEFT] = t;
	tData = tData->child[LEFT];
	depth++;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	//if (prefix == 134217728)	
	//cout << "created node at depth " << depth << "left = " << tData->child[LEFT] << "node = " << tData << endl;
      }
      break;
			
    case 1:
      flags[depth] = 2;
      if (Rib->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	Rib = Rib->parent;
	tData = tData->parent;
	depth--;
      }
      else	{
	Rib = Rib->child[RIGHT];
	cTempORTCNRData *t = new cTempORTCNRData (NULL,tData,NULL,NULL,-1,-1);
	iter++;
	tData->child[RIGHT] = t;
	tData = tData->child[RIGHT];
	depth++;
	prefix = prefix | (1 << (32-depth));
      }
      break;
		
    case 2:
      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      flags[depth] = 0;
      Rib = Rib->parent;
      tData = tData->parent;
      depth--;
    }
  }
  //cout << "ReadRib complete. Created " << iter << " nodes\n";
  cout << "Number of Unique NH so far = " << NHToIFaceMap.size() << endl;
  //exit(-1);
  return 0;
	
}

int cORTCUpdate::ORTCPass1()	{
  cTempORTCNRData* tData = TempData;
  cRIBdualNHNode* Rib = RIBdualNH->root;
  cTempORTCNRData* t = NULL;
  cRIBdualNHNode* n = NULL;
  unsigned int prefix = 0;
  vector <int> flags (33,0);
  vector <unsigned int> ImmediatePrefix (33,0);
  int iter = 0;
  int depth = 0;
  //cout << TempData << endl;
  while (tData != NULL)	{
    switch (flags[depth])	{
    case 0:
      //if (prefix == 134217728)
      //	cout << "node at depth " << depth << "left = " << tData->child[LEFT] << "node = " << tData << endl;
      //cout << prefix << "/" << depth << " " << tData << " " << Rib << endl;
      if (Rib == NULL)
	{
	  cout << "Rib NULL at " << prefix << "/" << depth << endl;
	  exit(1);
	}
      if (tData->IFace == -1)
	tData->IFace = tData->parent->IFace;
      else
	ImmediatePrefix[depth] = 1;
      if (tData->child[LEFT] == NULL && tData->child[RIGHT] != NULL)	{
	t = new cTempORTCNRData(NULL,tData,NULL,NULL,-1,(GetMask(ImmediatePrefix,depth)+1));
	tData->child[LEFT] = t;
	n = new cRIBdualNHNode(Rib,NULL,NULL);
	Rib->child[LEFT] = n;
	//if (prefix == 134217728)
	//	cout << "created node at depth " << depth << endl;
      }
      if (tData->child[RIGHT] == NULL && tData->child[LEFT] != NULL)	{
	t = new cTempORTCNRData(NULL,tData,NULL,NULL,-1,(GetMask(ImmediatePrefix,depth)+1));
	tData->child[RIGHT] = t;
	n = new cRIBdualNHNode(Rib,NULL,NULL);
	Rib->child[RIGHT] = n;
      }
      flags[depth] = 1;
      if (tData->child[LEFT] == NULL && tData->child[RIGHT] == NULL)	{
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	flags[depth] = 0;
	ImmediatePrefix[depth] = 0;
	depth--;
	iter++;
	tData = tData->parent;
	Rib = Rib->parent;
      }
      else	{
	tData = tData->child[LEFT];
	Rib = Rib->child[LEFT];
	depth++;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      }
      break;
			
    case 1:
      flags[depth] = 2;
      tData = tData->child[RIGHT];
      Rib = Rib->child[RIGHT];
      depth++;
      prefix = prefix | (1 << (32-depth));
      break;
		
    case 2:
      flags[depth] = 0;
      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      tData = tData->parent;
      Rib = Rib->parent;
      ImmediatePrefix[depth] = 0;
      depth--;
    }
  }
  //cout << "ORTC pass 1 complete\n";
  //cout << "leaves after pass 1 = " << iter << endl;
  //cout << "number of interfaces: " << NHToIFaceMap.size() << endl;
  return 0;
}	

int cORTCUpdate::ORTCPass2()	{
  int depth = 0;
  vector <int> flags (33,0);
  cTempORTCNRData* tData = TempData;
	
  while (tData != NULL)	{
    switch (flags[depth])	{
    case 0:
      if (tData->child[LEFT] != NULL)	{
	flags[depth] = 1;
	tData = tData->child[LEFT];
	depth++;
      }
      else	{
	tData->PotentialNHSet = CreateSet(tData->IFace);
	tData = tData->parent;
	depth--;
      }
      break;
		
    case 1:
      flags[depth] = 2;
      tData = tData->child[RIGHT];
      depth++;
      break;
		
    case 2:
      flags[depth] = 0;
      tData->PotentialNHSet = 
	HashOp (tData->child[LEFT]->PotentialNHSet,tData->child[RIGHT]->PotentialNHSet);
      tData = tData->parent;
      depth--;
    }
  }
  //cout << "ortc pass 2 complete\n";
  //cout << "intersections = " << inter << endl;
  return 0;
}

int cORTCUpdate::ORTCPass3()	{
  cRIBdualNHNode *n = RIBdualNH->root;
  cTempORTCNRData *t = TempData;	cTempORTCNRData* temp;
  cRIBdualNHNode *tRib;
  int depth = 0; 
  int m;
  unsigned int prefix = 0;
  vector <int> flags (33,0);
  vector <unsigned int> ImmediatePrefixAT (33,0);
  int iter = 0;	
  unsigned int currIF;

  while (t != NULL)	{
    switch (flags[depth])	{
    case 0:
      if (t->parent != NULL && InSet (t->parent->IFace, t->PotentialNHSet))	{
	if (n->GenNH != NULL_NEXTHOP)	{
	  n->GenNH = NULL_NEXTHOP;		// delete old nexthop
	  nodes_changed_ortc++;
	  n->DistanceToParent = -1;
	}
					
	t->IFace = t->parent->IFace;
	if (t->child[LEFT] == NULL && t->child[RIGHT] == NULL && t->ParentOffset != -1)	{
	  if (t->ParentOffset > GetMask (ImmediatePrefixAT,depth))	{
	    cRIBdualNHNode *r = n;
	    m = GetMask (ImmediatePrefixAT, depth);
	    for (int i = 0; i < m; i++)	
	      n = n->parent;
	    n->DistanceToParent = t->ParentOffset - m;
	    n = r;
	  }
	}
      }
      else	{
	iter++;		// number of entries
	
	/* this is old code being replaced by new code below
	t->IFace = choose (t->PotentialNHSet);
	ImmediatePrefixAT[depth] = 1;
	if (n->GenNH != NHToIFaceMap[t->IFace])	{
	  n->GenNH = NHToIFaceMap[t->IFace];		// assign new nexthop
	  n->DistanceToParent = -1;
	  nodes_changed_ortc++;
	}
	*/
	// start of new code
	if (n->GenNH == NULL_NEXTHOP)
	  {
	    t->IFace = choose (t->PotentialNHSet);
	    ImmediatePrefixAT[depth] = 1;
	    n->GenNH = NHToIFaceMap[t->IFace];// assign new nexthop
	    n->DistanceToParent = -1;
	    nodes_changed_ortc++;
	  }
	else
	  {
	    currIF = ConvertNHtoIFace(n->GenNH);
	    if (InSet (currIF, t->PotentialNHSet))
	      {
		t->IFace = currIF;
		n->DistanceToParent = -1;
		ImmediatePrefixAT[depth] = 1;
	      }
	    else
	      {
		t->IFace = choose (t->PotentialNHSet);
		ImmediatePrefixAT[depth] = 1;
		n->GenNH = NHToIFaceMap[t->IFace];
		n->DistanceToParent = -1;
		nodes_changed_ortc++;
	      }
	          
	  }
	// end of new code
	if (n->child[LEFT] == NULL && n->child[RIGHT] == NULL && n->BestOrigNH == NULL_NEXTHOP)
	  n->DistanceToParent = t->ParentOffset;
	else
	  n->DistanceToParent = -1;
      }
				
      if (t->child[LEFT] == NULL && t->child[RIGHT] == NULL)	{
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);     // unset prefix
	flags[depth] = 0;	
	ImmediatePrefixAT[depth] = 0;	
	depth--;
	temp = t;
	tRib = n;
	t = t->parent;
	n = n->parent;
	if (tRib->BestOrigNH == NULL_NEXTHOP && tRib->GenNH == NULL_NEXTHOP)	{
	  if (n->child[LEFT] == tRib)
	    n->child[LEFT] = NULL;
	  else
	    n->child[RIGHT] = NULL;
	  delete tRib;
	}
	delete temp;
      }
      else	{
	flags[depth] = 1;
	t = t->child[LEFT];
	n = n->child[LEFT];
	depth++;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      }
      break;

    case 1:
      flags[depth] = 2;
      depth++;
      t = t->child[RIGHT];
      n = n->child[RIGHT];
      prefix = prefix | (1 << (32-depth));	// set bit
      break;
	
    case 2:
      flags[depth] = 0;
      ImmediatePrefixAT[depth] = 0;	
      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      depth--;
      temp = t;
      t = t->parent;
      n = n->parent;
      delete temp;
    }
  }
  //cout << iter << endl;
  return iter;
}
	

/************************************************************************************************************************
ORTC based Update Algorithms
************************************************************************************************************************/

int cORTCUpdate::Update (unsigned int Prefix, int mask, unsigned int OldNH, int MessageType)	{
  effective_updates++;
  effective_updates_by_prefix[mask]++;
  if (MessageType == 0)	// withdraw
    ORTCWithdraw (Prefix, mask, OldNH);
  else if (MessageType == 1)	// Announce replace
    ORTCAnnounce (Prefix, mask, OldNH, 1);
  else if (MessageType == 2)
    ORTCAnnounce (Prefix, mask, OldNH, 2);
  else	{
    cout << "&^%$#(*&#%! Unknown Message type" << endl;
    exit(1);
  }
  return 0;
}

int cORTCUpdate::ORTCWithdraw (unsigned int Prefix, int mask, unsigned int OldNH)	{
  cRIBdualNHNode *next = RIBdualNH->root, *P = NULL, *I = NULL, *N = NULL;
  int nbit = 32;
  int tmask = mask;
  int PMask = 0, IMask = 0;
  int bit;
  unsigned int X;				// Nexthop to restore in aggregated tree
  int Step1 = 1;
  unsigned int old_gen_nh;
  unsigned int prefix = Prefix;
  unsigned int R;				// Nexthop to remove from aggregated tree
  while (tmask > 0)	{			// preorder tree traversal to note down P and I

    if (next->BestOrigNH != NULL_NEXTHOP)	{
      PMask = 32 - nbit;	
      P = next;
    }
    if (next->GenNH != NULL_NEXTHOP)	{	
      I = next;	
      IMask = 32 - nbit;
    }		
    nbit--;
				
    bit = (Prefix >> nbit) % 2;
    next = next->child[bit];
    tmask--;
		
  }
  N = next;
  old_gen_nh = N->GenNH;
  // At this stage we have P and I
  //cout << Prefix << "/" << mask << endl;
  //cout << P->BestOrigNH << " " << PMask << endl;
  //cout << I->GenNH << " " << IMask << endl;
  //cout << N->GenNH << endl;
  if (next->GenNH != NULL_NEXTHOP)	{
    if (next->GenNH == OldNH)	{		// OldNH was aggregated here
      X = OldNH;
      R = P->BestOrigNH;
      N->GenNH = NULL_NEXTHOP;
    }
    else	{
      I = N;
      Step1 = 0;
      IMask = mask; 
			
    }
  }
  else	{
    X = I->GenNH;
  }
  //cout << "Step 1 is: " << Step1 << endl;
	
  if (P->BestOrigNH != I->GenNH && Step1 != 0)	{
    N->GenNH = P->BestOrigNH;
    R = N->GenNH;
    //cout << "replacing N with P" << endl;
    N->DistanceToParent = mask - PMask;
		
    //cout << "deaggregate set " << N->GenNH << " " << mask << " N" << endl;
    //cout << N->GenNH << endl;
  }
  else if (PMask < IMask && Step1 != 0)	{
    R = I->GenNH;
    //cout << "Withdraw of " << Prefix << "/" << mask << " " << N->BestOrigNH << " " << N->GenNH << "replacing " << OldNH << endl;
    //cout << "P is " << P->BestOrigNH << endl;
    //cout << "I is " << I->GenNH << endl;
    //if (Bool == 1)	{
    I->DistanceToParent = IMask - PMask;
    //cout << "deaggregate set " << I->GenNH << " " << IMask << " I" << endl;
    //}
    if (N->child[LEFT] == NULL && N->child[RIGHT] == NULL)	{
      if (N->parent->child[LEFT] == N)
	N->parent->child[LEFT] = NULL;
      else if (N->parent->child[RIGHT] == N)
	N->parent->child[RIGHT] = NULL;
      delete N;
      return 0;
    }
  }
  //cout << "Withdraw of " << Prefix << "/" << mask << " " << N->BestOrigNH << " " << N->GenNH << "replacing " << OldNH << endl;
  //cout << "P is " << P->BestOrigNH << endl;
  //cout << "I is " << I->GenNH << endl;

  if (old_gen_nh != N->GenNH)	{
    nodes_changed_smalta++;
    nodes_changed_by_prefix[mask]++;
  }
  if (P->BestOrigNH != X && Step1 != 0)	{			// deaggregate immediate descendents
	
    next = N;
    cRIBdualNHNode *terminate = N->parent;
    vector <int> flags (33,0);
    int depth = mask;
    int TraverseDown = 0;
    while (next != terminate)	{
      switch (flags[depth])	{
      case 0:
	if (next != N && (next->BestOrigNH != NULL_NEXTHOP || next->GenNH != NULL_NEXTHOP))	{
	  //cout << "considering " << next->BestOrigNH << " " << next->GenNH << " " << prefix << "/" << depth << endl;
	  if (next->GenNH == R || next->BestOrigNH == R)	{
	    if (next->GenNH == R)	{
	      //cout << "removing " << next->GenNH << " " << depth << endl;
	      next->GenNH = NULL_NEXTHOP;
	      // A removal of a node is an update
	      nodes_changed_smalta++;
	      nodes_changed_by_prefix[mask]++;
	      //if (next->BestOrigNH == NULL_NEXTHOP)	
	      next->DistanceToParent = -1;
	    }
	    else if (next->BestOrigNH == R && next->GenNH == NULL_NEXTHOP)	{
	      //cout << "Traverse down = 1\n";
	      TraverseDown = 1;
	    }
	  }
	  else	{
	    if (next->BestOrigNH == X && next->GenNH == NULL_NEXTHOP)	{
	      //cout << "restoring " << next->BestOrigNH << " " << depth << endl;	
	      next->GenNH = next->BestOrigNH;
	      // announce to the fib hence a download
	      nodes_changed_smalta++;
	      nodes_changed_by_prefix[mask]++;
	    }
	    else if (next->BestOrigNH != X && next->GenNH == NULL_NEXTHOP)	{
	      //cout << "Traverse Down set" << endl;
	      TraverseDown = 1;
	    }
	  }
	  if (TraverseDown == 0)	{
	    flags[depth] = 2;
	    break;
	  }	
	  else
	    TraverseDown = 0;
	}
	flags[depth] = 1;
	if (next->child[LEFT] == NULL)	{
	  flags[depth] = 2;
	  if (next->child[RIGHT] == NULL)	{
	    flags[depth] = 0;
	    next = next->parent;
	    prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	    depth--;
	  }
	  else	{
	    next = next->child[RIGHT];
	    depth++;
	    prefix = prefix | (1 << (32-depth));
	  }
	}
	else	{
	  next = next->child[LEFT];
	  depth++;
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	}
	break;
	
      case 1:
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  next = next->parent;
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	  prefix = prefix | (1 << (32-depth));
	}
	break;
		
      case 2:
	flags[depth] = 0;
	next = next->parent;	
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	depth--;
      }
    }
		
  }
	
  // parse the deaggregates to reclaim space occupied by N before withdraw in favor of P

  next = N;
  int depth = mask;
  cRIBdualNHNode* terminate = N->parent;
  vector <int> flags (33,0);
  prefix = Prefix;
  while (next != terminate)	{
    switch (flags[depth])	{
    case 0:
      if (next != N && next->GenNH != NULL_NEXTHOP)	{
	if (next->DistanceToParent == (depth - mask))	{
	  //cout << "deaggregated encountered\n";
	  if (next->GenNH != OldNH)			
	    cout << "wrong deaggregate " << prefix << "/" << depth << " " << next->GenNH << " " << next->BestOrigNH << "distance " << next->DistanceToParent << endl;
	  next->DistanceToParent = (depth - PMask);
	  next->GenNH = P->BestOrigNH;
	  // a change to the nexthop corresponds to a change in FIB
	  nodes_changed_smalta++;
	  nodes_changed_by_prefix[mask]++;
	  // Now restore the appropriate immediate descendents of next
	  {
	    int rdepth = depth;
	    cRIBdualNHNode *rN = next;
	    cRIBdualNHNode *rterminate = rN->parent;
	    cRIBdualNHNode *rnext = rN;
	    vector <int> rflags (33,0);
	    int TraverseDown = 0;
	    while (rnext != rterminate)	{
	      switch (rflags[rdepth])	{
	      case 0:
		if ((rnext != rN && rnext->GenNH != NULL_NEXTHOP) ||
		    (rnext->BestOrigNH != NULL_NEXTHOP))	{
		  if ((rnext->GenNH == rN->GenNH) || (rnext->BestOrigNH == rN->GenNH))	{
		    if (rnext->GenNH == rN->GenNH)	{
		      rnext->GenNH = NULL_NEXTHOP;
		      // a withdraw from fib
		      nodes_changed_smalta++;
		      nodes_changed_by_prefix[mask]++;
		      if (rnext->BestOrigNH == NULL_NEXTHOP)	
			rnext->DistanceToParent = -1;
		    }
		    else if (rnext->BestOrigNH == rN->GenNH && rnext->GenNH == NULL_NEXTHOP)
		      TraverseDown = 1;
		  }
		  else	{
		    if ((rnext->BestOrigNH == OldNH) && (rnext->GenNH == NULL_NEXTHOP))	{
		      rnext->GenNH = rnext->BestOrigNH;
		      // an announce to the fib
		      nodes_changed_smalta++;
		      nodes_changed_by_prefix[mask]++;
		    }
		    else if ((rnext->BestOrigNH != OldNH) && (rnext->GenNH == NULL_NEXTHOP))
		      TraverseDown = 1;
		  }
		  if (TraverseDown == 0)	{
		    rflags[rdepth] = 2;
		    break;	
		  }
		  else
		    TraverseDown = 0;
		}
		rflags[rdepth] = 1;
		if (rnext->child[LEFT] == NULL)	{
		  rflags[rdepth] = 2;
		  if (rnext->child[RIGHT] == NULL)	{
		    rflags[rdepth] = 0;
		    rnext = rnext->parent;
		    rdepth--;
		  }
		  else	{
		    rnext = rnext->child[RIGHT];
		    rdepth++;
		  }
		}
		else	{
		  rnext = rnext->child[LEFT];
		  rdepth++;
		}
		break;
			
	      case 1:
		rflags[rdepth] = 2;
		if (rnext->child[RIGHT] == NULL)	{
		  rflags[rdepth] = 0;
		  rnext = rnext->parent;
		  rdepth--;
		}
		else	{
		  rnext = rnext->child[RIGHT];
		  rdepth++;
		}
		break;

	      case 2:
		rflags[rdepth] = 0;
		rnext = rnext->parent;
		rdepth--;
	      }	
	    }
	  }
	}
      }
      flags[depth] = 1;
      if (next->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  next = next->parent;
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	  prefix = prefix | (1 << (32-depth));
						
	}
      }
      else	{
	next = next->child[LEFT];
	depth++;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      }
      break;
			
    case 1:
      flags[depth] = 2;
      if (next->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = next->parent;
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	depth--;
      }
      else	{
	next = next->child[RIGHT];
	depth++;
	prefix = prefix | (1 << (32-depth));
      }	
      break;
				
    case 2:
      flags[depth] = 0;
      next = next->parent;
      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
      depth--;
    }
  }
  //cout << N->GenNH << endl;
	
  return 0;
}

	
int cORTCUpdate::ORTCAnnounce (unsigned int Prefix, int mask, unsigned int OldNH, int SubMessage)	{
  cRIBdualNHNode *next = RIBdualNH->root;
  cRIBdualNHNode *P = NULL, *I = NULL, *N = NULL, *terminate = NULL;
  int nbit = 32;
  int tmask = mask;
  int PMask = 0, IMask = 0;
  int bit;
  unsigned int R;
  unsigned int X;
  unsigned int old_gen_nh;
  unsigned int prefix = Prefix;
  int Step1 = 1;
  while (tmask > 0)	{
    if (next->GenNH != NULL_NEXTHOP)	{
      I = next;
      IMask = 32 - nbit;
    }
    if (next->BestOrigNH != NULL_NEXTHOP)	{
      P = next;
      PMask = 32 - nbit;
    }

    nbit--;
				
    bit = (Prefix >> nbit) % 2;
    next = next->child[bit];
    tmask--;
  }
  N = next;
  old_gen_nh = N->GenNH;
  //cout << "Announce of " << Prefix << "/" << mask << " " << N->BestOrigNH << " " << N->GenNH << " replacing " << OldNH << " submtype" << SubMessage << endl;
  N->DistanceToParent = -1;
  terminate = N->parent;
	
  if (OldNH != NULL_NEXTHOP)	{
    P = N;
    //PMask = 32 - nbit;   // old value
    PMask = mask;
  }

  if (N->GenNH == NULL_NEXTHOP)	{
    if (N->BestOrigNH != I->GenNH)	{
      N->GenNH = N->BestOrigNH;
      R = N->BestOrigNH;
      X = I->GenNH;
    }
    else
      Step1 = 0;
  }
  else	{
    if (OldNH == NULL_NEXTHOP || OldNH == N->GenNH)	{
      X = N->GenNH;
      R = N->BestOrigNH;
      if (N->BestOrigNH == I->GenNH)
	N->GenNH = NULL_NEXTHOP;
      else
	N->GenNH = N->BestOrigNH;
    }
    else
      Step1 = 0;
  }
	
	

  //cout << "I is " << I->GenNH << endl;
  //cout << "P is " << P->BestOrigNH << endl;
  /*
    if (SubMessage == 1)	{
    PMask = 32 - nbit;
    P = N;
    if (N->GenNH == NULL_NEXTHOP)	{
    //cout << "N->GenNH is NULL\n";
    X = I->GenNH;
    if (N->BestOrigNH == X)
    Step1 = 0;
    else	{
    R = N->GenNH;
    N->GenNH = N->BestOrigNH;
    }
    }				
    else	{
    //cout << "N->GenNH is " << N->GenNH << endl;
    X = OldNH;
    if (OldNH == N->GenNH)	{
    if (N->BestOrigNH == I->GenNH)	{	
    //Step1 = 0;
    N->GenNH = NULL_NEXTHOP;
    R = NULL_NEXTHOP;
    }
    else	{
    N->GenNH = N->BestOrigNH;
    R = N->GenNH;
    }
    }
    else	
    Step1 = 0;
    }
    }
    else if (SubMessage == 2)	{
    if (N->GenNH != NULL_NEXTHOP)	{	
    X = N->GenNH; 
    N->GenNH = N->BestOrigNH;
    R = N->GenNH;
    if (N->BestOrigNH == I->GenNH)	{
    R = NULL_NEXTHOP;
    N->GenNH = NULL_NEXTHOP;
    }
    }
    else	{
    X = I->GenNH;
    if (N->BestOrigNH == X)	{
    Step1 = 0;
    N->GenNH = NULL_NEXTHOP;
    }
    else	{
    N->GenNH = N->BestOrigNH;
    R = N->GenNH;
    }
    }
    }
  */
  //cout << "After update N is " << N->GenNH << endl;
  if (old_gen_nh != N->GenNH)	{
    nodes_changed_smalta++;
    nodes_changed_by_prefix[mask]++;
  }
  if (Step1 != 0)	{
    next = N;
    vector <int> flags (33,0);
    int depth = mask;
    int TraverseDown = 0;
    while (next != terminate)	{
      switch (flags[depth])	{
      case 0:
	if (next != N && (next->GenNH != NULL_NEXTHOP || next->BestOrigNH != NULL_NEXTHOP))	{
	  //cout << "Considering orig " << next->BestOrigNH << " gen " << next->GenNH << 
	  //" " << depth << endl; 
	  if ((next->GenNH == R || next->BestOrigNH == R) && R != NULL_NEXTHOP)	{
	    //cout << "removing " << next->GenNH << " " << depth << endl;
	    if (next->GenNH == R)	{
	      next->GenNH = NULL_NEXTHOP;
	      // a withdraw to the fib
	      nodes_changed_smalta++;
	      nodes_changed_by_prefix[mask]++;
	      if (next->BestOrigNH == NULL_NEXTHOP)	
		next->DistanceToParent = -1;
	    }
	    else if (next->BestOrigNH == R && next->GenNH == NULL_NEXTHOP)	
	      TraverseDown = 1;
	  }
	  else	{
	    if (next->GenNH == NULL_NEXTHOP && next->BestOrigNH == X)	{
	      //cout << "restoring " << next->BestOrigNH << endl;	
	      next->GenNH = next->BestOrigNH;
	      // an announce to teh fib
	      nodes_changed_smalta++;
	      nodes_changed_by_prefix[mask]++;
	    }
	    else if (next->BestOrigNH != X && next->GenNH == NULL_NEXTHOP) {	
	      TraverseDown = 1;
	      //cout << "Traverse down set \n";
	    }
	  }
	  if (TraverseDown == 0)	{
	    flags[depth] = 2;
	    break;
	  }	
	  else
	    TraverseDown = 0;
	}
	flags[depth] = 1;
	if (next->child[LEFT] == NULL)	{
	  flags[depth] = 2;
	  if (next->child[RIGHT] == NULL)	{
	    flags[depth] = 0;
	    next = next->parent;
	    prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	    depth--;
	  }
	  else	{
	    next = next->child[RIGHT];
	    depth++;
	    prefix = prefix | (1 << (32-depth));
	  }
	}
	else	{
	  next = next->child[LEFT];
	  depth++;	
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	}
	break;
	
      case 1:
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  next = next->parent;
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	  prefix = prefix | (1 << (32-depth));
	}
	break;
		
      case 2:
	flags[depth] = 0;
	next = next->parent;	
	prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	depth--;
      }
    }
  }
	
  next = N;
  int depth = mask;
  terminate = N->parent;
  vector <int> flags (33,0);
  int TraverseDown = 0;
  unsigned int V;
  if (OldNH == NULL_NEXTHOP)
    V = P->BestOrigNH;	
  else
    V = OldNH;
  prefix = Prefix;
  while (next != terminate)	{
    switch (flags[depth])	{
    case 0:
      if (next != N && next->GenNH != NULL_NEXTHOP)	{
	if (next->DistanceToParent == (depth - PMask))	{
	  //cout << "deaggregate encountered " << next->GenNH << " " << next->BestOrigNH << endl;
	  if (next->GenNH != V)//  {
	    cout << "wrong deaggregate " << next->GenNH << " " << next->BestOrigNH  << endl;
	  //	goto skip;
	  //}
	  next->DistanceToParent = (depth - mask);	
	  next->GenNH = N->BestOrigNH;
	  // a change in the nexthop
	  nodes_changed_smalta++;
	  nodes_changed_by_prefix[mask]++;
						
	  // Now restore the appropriate immediate descendents of next
	  {
	    int rdepth = depth;
	    cRIBdualNHNode *rN = next;
	    cRIBdualNHNode *rnext = rN;
	    cRIBdualNHNode *rterminate = rN->parent;
	    vector <int> rflags (33,0);
	    while (rnext != rterminate)	{
	      switch (rflags[rdepth])	{
	      case 0:
		if ((rnext != rN && rnext->GenNH != NULL_NEXTHOP) ||
		    (rnext->BestOrigNH != NULL_NEXTHOP) )	{
		  if ((rnext->GenNH == rN->GenNH) || (rnext->BestOrigNH == rN->GenNH))	{
		    if (rnext->GenNH == rN->GenNH)	{
		      rnext->GenNH = NULL_NEXTHOP;
		      // a withdraw
		      nodes_changed_smalta++;
		      nodes_changed_by_prefix[mask]++;
		      if (rnext->BestOrigNH == NULL_NEXTHOP)	
			rnext->DistanceToParent = -1;
		    }
		    else if (rnext->BestOrigNH == N->GenNH && rnext->GenNH == NULL_NEXTHOP)	
		      TraverseDown = 1;
		  }
		  else	{
		    if (rnext->GenNH == NULL_NEXTHOP && rnext->BestOrigNH == V)	{
		      rnext->GenNH = rnext->BestOrigNH;
		      // an announce
		      nodes_changed_smalta++;	
		      nodes_changed_by_prefix[mask]++;
		    }
		    else if (rnext->BestOrigNH != V && rnext->GenNH == NULL_NEXTHOP) 	
		      TraverseDown = 1;
								
		  }
										
		  if (TraverseDown == 0)	{
		    rflags[rdepth] = 2;
		    break;	
		  }
		  else
		    TraverseDown = 0;
		}
		rflags[rdepth] = 1;
		if (rnext->child[LEFT] == NULL)	{
		  rflags[rdepth] = 2;
		  if (rnext->child[RIGHT] == NULL)	{
		    rflags[rdepth] = 0;
		    rnext = rnext->parent;
		    rdepth--;
		  }
		  else	{
		    rnext = rnext->child[RIGHT];
		    rdepth++;
		  }
		}
		else	{
		  rnext = rnext->child[LEFT];
		  rdepth++;
		}
		break;
			
	      case 1:
		rflags[rdepth] = 2;
		if (rnext->child[RIGHT] == NULL)	{
		  rflags[rdepth] = 0;
		  rnext = rnext->parent;
		  rdepth--;
		}
		else	{
		  rnext = rnext->child[RIGHT];
		  rdepth++;
		}
		break;

	      case 2:
		rflags[rdepth] = 0;
		rnext = rnext->parent;
		rdepth--;
	      }	
	    }
	  }
	}
      }
      flags[depth] = 1;
      if (next->child[LEFT] == NULL)	{
	flags[depth] = 2;
	if (next->child[RIGHT] == NULL)	{
	  flags[depth] = 0;
	  next = next->parent;
	  depth--;
	}
	else	{
	  next = next->child[RIGHT];
	  depth++;
	}
      }
      else	{
	next = next->child[LEFT];
	depth++;
      }
      break;
			
    case 1:
      flags[depth] = 2;
      if (next->child[RIGHT] == NULL)	{
	flags[depth] = 0;
	next = next->parent;
	depth--;
      }
      else	{
	next = next->child[RIGHT];
	depth++;
      }	
      break;
				
    case 2:
      flags[depth] = 0;
      next = next->parent;
      depth--;
    }
  }
	
  return 0;


}
					
	
