/***************************************************************************
 *   Copyright (C) 2010 by Ahsan Tariq   *
 *   ahsan@mpi-sws.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <vector>
#include <string>
#include <cstring>
#include <climits>

#define NULL_NEXTHOP UINT_MAX
#define DROP 0
#define LEFT 0
#define RIGHT 1

class cRIB;
class cRIBNode;
class cRIBToFIBDownloader;
class cRIBdualNHNode;
class cRIBdualNH;

class cTempORTCNRData	
{
 public:
  unsigned int* PotentialNHSet;
  cTempORTCNRData *parent;
  cTempORTCNRData *child[2];
  int IFace;
  int ParentOffset;			// Used for assigning Distance to Parent 
  
  cTempORTCNRData 
    (unsigned int* nh, cTempORTCNRData *p, cTempORTCNRData *l, cTempORTCNRData *r, 
     int i, int ppref)	
    {
      parent = p;
      child[LEFT] = l;
      child[RIGHT] = r;
      PotentialNHSet = nh;
      IFace = i;
      ParentOffset = ppref;
    }
  ~cTempORTCNRData()	
    { 
      delete PotentialNHSet; 	
    }
};

class cZhangAgg	
{
 public:
  int Apply();
  void Update (unsigned int Prefix, int Mask, int Message);
  cZhangAgg (cRIB *pRIB, cRIBToFIBDownloader *pRIBToFIBDownloader, int Level)
    { 
      RIB = pRIB;	
      RIBToFIBDownloader = pRIBToFIBDownloader; 
      AggLevel = Level;	
    }
  unsigned int GetNodesChangedLevel1 ()	{return NodesLevel1;}
  unsigned int GetNodesChangedLevel2 ()	{return NodesLevel2;}
 
 private:
  cRIB *RIB;
  cRIBToFIBDownloader *RIBToFIBDownloader;
  int AggLevel;
  void ApplyLevel1();
  void ApplyLevel2();
  unsigned int GetImmAncestorNH (std::vector<unsigned int>& ImmAnc, int d);
  unsigned int NodesLevel1;
  unsigned int NodesLevel2;
  
  void reAggregate(cRIBNode*,int);
  void deAggregate(cRIBNode*,cRIBNode*,int);
};
		

class cSMALTA	
{
 public:
  cSMALTA (cRIB *pRIB, cRIBToFIBDownloader *pRIBToFIBDownloader) 
    {
      RIB = pRIB;
      RIBToFIBDownloader = pRIBToFIBDownloader;
      NodesChanged = 0;
      FirstTimeORTC = 0;
      NodesChanged = 0;
      TempORTCNRData = new cTempORTCNRData (NULL, NULL, NULL, NULL, 0, -1);
    }
  int SMALTAUpdate (unsigned int Prefix, int mask, unsigned int OldNH, int MessageType);
  unsigned int GetNodesChanged () { return NodesChanged; }
  int ApplyORTCNR ();	
  
 private:
  cRIB *RIB;
  int FirstTimeORTC;
  cRIBToFIBDownloader *RIBToFIBDownloader;
  unsigned int NodesChanged;
  int SMALTAWithdraw (unsigned int Prefix, int mask, unsigned int OldNH);
  int SMALTAAnnounce (unsigned int Prefix, int mask, unsigned int OldNH, int subMessage);
  int intersections;
  int leaves;
  std::vector <unsigned int> NHtoIFaceMap;
  int ConvertNHtoIFace (unsigned int NH);
  cTempORTCNRData* TempORTCNRData;
  unsigned int* HashOp(unsigned int* l,unsigned int* r);
  unsigned int* CreateSet(int iface);
  int InSet (int i, unsigned int* a);
  int choose (unsigned int* a);
  int GetMask(std::vector<int>& ImmPref, int d);
  int ReadRIB ();
  int ORTCNRPass1 ();
  int ORTCNRPass2 ();
  int ORTCNRPass3 ();
}; 

class cORTCUpdate	
{
 public:
  cORTCUpdate(cRIBdualNH* pRIB, cRIBToFIBDownloader *pDownloader)	
    {
      RIBdualNH = pRIB;	RIBToFIBDownloader = pDownloader;
    }
  //  void attachNewRIB(cRIBdualNH *pRIB) {RIBdualNH = pRIB;}
  int Update (unsigned int Prefix, int mask, unsigned int OldNH, int MessageType);
  int ApplyORTC ();
  unsigned int GetORTCRIBtoFIBdl ()	{ return nodes_changed_ortc; }
  unsigned int GetSMALTARIBtoFIBdl ()	{ return nodes_changed_smalta; }
  unsigned int GetEffectiveUpdates ()	{ return effective_updates;	}
  unsigned int* GetEffectiveUpdatesByPrefix ()	{ return effective_updates_by_prefix;	}
  unsigned int* GetNodesChangedByPrefix ()	{ return nodes_changed_by_prefix;	}
  
 private:
  unsigned int effective_updates;
  unsigned int effective_updates_by_prefix[33];
  unsigned int nodes_changed_by_prefix[33];
  unsigned int nodes_changed_ortc;
  unsigned int nodes_changed_smalta;
  cRIBdualNH *RIBdualNH;
  cRIBToFIBDownloader *RIBToFIBDownloader;
  int ConvertNHtoIFace (unsigned int NH);
  cTempORTCNRData *TempData;
  std::vector <unsigned int> NHToIFaceMap;
  unsigned int* HashOp (unsigned int* l, unsigned int *r);
  unsigned int* CreateSet (int iface);
  int InSet (int i, unsigned int* a);
  int choose (unsigned int* a);
  int GetMask (std::vector<unsigned int>& ImmPref, int d);
  int ReadRIB();
  int ORTCPass1 ();
  int ORTCPass2 ();
  int ORTCPass3 ();
  int ORTCWithdraw(unsigned int Prefix, int mask, unsigned int OldNH);
  int ORTCAnnounce(unsigned int Prefix, int mask, unsigned int OldNH, int SubMessage);
};


		
