/***************************************************************************
 *   Copyright (C) 2010 by Ahsan Tariq   *
 *   ahsan@mpi-sws.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <vector>
#include <string>
#include <cstring>
#include <climits>

#define NULL_NEXTHOP UINT_MAX
#define DROP 0
#define LEFT 0
#define RIGHT 1

class cSMALTA;
class cZhangAgg;
class cORTCUpdate;

class cPolicy 
{
 private:
  int Policy;
		
  friend class cRIBNode;
 public:
  cPolicy(int policy)	{ Policy = policy; }
  int GetPolicy()	{ return Policy; }
};

struct sNHASLenPair	
{
  unsigned int NH;
  int ASLen;
};

struct sRoutes	
{
  std::vector<sNHASLenPair> NHASLenPair;
  unsigned int BestNH;
  int BestASLen;
};

struct sRIBFlags	
{
  int INFIB;
};

struct sSMALTAFlags	
{
  int Original;
  int DistanceToParent;
};	

struct sData	
{
  sRoutes Routes;
  sRIBFlags RIBFlags;
  sSMALTAFlags SMALTAFlags;
};

class cRIBNode	
{
 public:
  sData Data;
  cRIBNode *parent;
  cRIBNode *child[2];
  int SetRoutes (unsigned int NH, int ASLen, int SMALTAEnabled, int ZhangAggEnabled);
  cRIBNode () {}
  cPolicy *pol;
  int UnSetRoutes (unsigned int peerIP, int SMALTAEnabled, int ZhangAggEnabled);
  cRIBNode (int INFIB, int Original, cRIBNode *par, cRIBNode *lchild, cRIBNode *rchild, cPolicy *p);	
  //int SMALTAEnabled;	// is set when ORTC is run before incorporating updates.
	
		
};

class cRIB	
{
 private:
  cRIBNode *root;
  int RIBAnnounce (unsigned int prefix, int mask, unsigned int nexthop, int ASLength);	// 
  int RIBWithdraw (unsigned int prefix, int mask, unsigned int peerIP);			// nexthop == peerIP
  cPolicy *pol;
  int SMALTAEnabled;
  int ZhangAggEnabled;
  cSMALTA *SMALTA;
  cZhangAgg *ZhangAgg;
  unsigned int NodesChanged;

 public:
  unsigned int RIBUpdate (unsigned int Prefix, int mask, unsigned int nexthop, int ASLength, int MessageType);
  int GetOriginalTable (std::string* FileName); 	
  void SetSMALTA (cSMALTA *p)	{	SMALTA = p;	}	
  friend class cSMALTA;
  friend class cZhangAgg;
  cRIB (cPolicy *p, int s);	
  void SetSMALTAEnabled (int s)	{ SMALTAEnabled = s; ZhangAggEnabled = 0; }
  void SetZhangAgg (cZhangAgg* z)	{ ZhangAgg = z;}
  void SetZhangAggEnabled (int s)	{ ZhangAggEnabled = s; SMALTAEnabled = 0; }
  int GetAggregatedTable (std::string *file);
}; 

class cRIBdualNHNode	
{
 public:
  cRIBdualNHNode *parent;
  cRIBdualNHNode *child[2];
  std::vector <sNHASLenPair> NHASLenPair;
  unsigned int BestOrigNH;
  int BestASLen;
  unsigned int GenNH;
  int DistanceToParent;
  cRIBdualNHNode (cRIBdualNHNode *par, cRIBdualNHNode *lchild, cRIBdualNHNode *rchild);
  int SetRoutes (unsigned int NH, int ASLen);
  int UnSetRoutes (unsigned int peerIP, int ORTCUpdateEnabled);
};



class cRIBdualNH	
{
 private:
  cRIBdualNHNode	*root;
  int RIBAnnounce (unsigned int Prefix, int mask, unsigned int nexthop, int ASLength);	// 
  int RIBWithdraw (unsigned int prefix, int mask, unsigned int peerIP);			// nexthop == peerIP
  int ORTCUpdateEnabled;
  cORTCUpdate *ORTCUpdate;
		
 public:
  std::vector<unsigned int> debugVector;
  friend class cORTCUpdate;
  cRIBdualNH()	{ORTCUpdateEnabled = 0; root = new cRIBdualNHNode (NULL,NULL,NULL);}
  void SetORTCUpdateEnabled (int s)	{	ORTCUpdateEnabled = s;	}
  void SetORTCUpdate (cORTCUpdate *p)	{	ORTCUpdate = p;	}
  int GetAggregatedTable (std::string *file = NULL);
  int GetOriginalTable (std::string *file = NULL);
  int Update (unsigned int Prefix, int mask, unsigned int nexthop, int ASLength, int MessageType);
		
};

