/***************************************************************************
 *   Copyright (C) 2010 by Ahsan Tariq   *
 *   ahsan@mpi-sws.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cstring>
#include <cstdlib>

#include "RouteProcessor.h"
#include "RIB.h"

using namespace std;

int h = 0;

cRouteProcessor::cRouteProcessor(string& RTFile, string& updateFile)
{
  numEntriesReadFromRT = 0;
  numUpdatesRead = 0;

  RTFilename = RTFile;
  updateFilename = updateFile;
}

/*
int cRouteProcessor::SetRoutingTable (std::string *RT)
{
  RoutingTable.open ((*RT).c_str());
  if (!RoutingTable)	
    return 1;
  else 
    return 0;
}

int cRouteProcessor::SetUpdateFile (std::string *UPFile)
{
  UpdateFile.open ((*UPFile).c_str());
  if (!UpdateFile)
    return 1;
  else
    return 0;
}

void cRouteProcessor::ResetRoutingTable()
{
  RoutingTable.clear();
  RoutingTable.close();
}
*/

// to give an opportunity to the client to start reading
// the update trace from the beginning
// (after reading so many updates)
void cRouteProcessor::resetUpdateFile()
{
  updateFileStream.clear();
  updateFileStream.close();
}

unsigned int cRouteProcessor::ReadUpdateFile(cRIB& RIB, int numUpdToBeRead)
{
  if(!updateFileStream.is_open())
    updateFileStream.open(updateFilename.c_str());

  if(!updateFileStream.good())
    {
      // file could not be opened for reading
      cout << "Error reading update trace file..." << endl;
      exit(-1);
    }

  // to keep track of how many entries be read in one go
  int numUpdatesRead = 0;
  
  // the following loop reads and processes one line in each iteration
  while((updateFileStream.peek() != EOF) && updateFileStream.good())
    {
      string oneLine; // the string to be read from the file
      
      // variables to be read from each line of the file
      unsigned int prefix;
      int maskLength;
      unsigned int nextHop;
      int ASLength;
      char verticalBar;
      char updateType; // Announce (A) or Withdraw (W)
      
      if(!getline(updateFileStream, oneLine))
	{
	  cout << "couldn't read from the file. badly formatted???" << endl;
	  updateFileStream.close();
	  exit(-1);
	}
      
      if(!oneLine.empty())
	{
	  numUpdatesRead++; // count only a non-blank line
	  
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << oneLine;
	  ss >> updateType >> verticalBar >> prefix >> verticalBar
	     >> maskLength >> verticalBar >> nextHop 
	     >> verticalBar >> ASLength;
	  
	  // TODO: check if the line format was correct. if not raise error
	  if(updateType == 'W')
	    RIB.RIBUpdate(prefix, maskLength, nextHop, ASLength, 0);
	  else if(updateType == 'A')
	    RIB.RIBUpdate(prefix, maskLength, nextHop, ASLength, 1);
	  else
	    {
	      cout << "Unknown Update Type" << endl;
	      exit(-1);
	    }
	}
      
      if(numUpdatesRead == numUpdToBeRead)
	return numUpdToBeRead;
    }
  
  resetUpdateFile(); // reset flags and close the file

  return numUpdatesRead;
}

unsigned int cRouteProcessor::ReadUpdateFile(cRIBdualNH& RIB, int numUpdToBeRead)
{
  if(!updateFileStream.is_open())
    updateFileStream.open(updateFilename.c_str());

  if(!updateFileStream.good())
    {
      // file could not be opened for reading
      cout << "Error reading update trace file..." << endl;
      exit(-1);
    }

  // to keep track of how many entries be read in one go
  int numUpdatesRead = 0;
  
  // the following loop reads and processes one line in each iteration
  while((updateFileStream.peek() != EOF) && updateFileStream.good())
    {
      string oneLine; // the string to be read from the file
      
      // variables to be read from each line of the file
      unsigned int prefix;
      int maskLength;
      unsigned int nextHop;
      int ASLength;
      char verticalBar;
      char updateType; // Announce (A) or Withdraw (W)
      
      if(!getline(updateFileStream, oneLine))
	{
	  cout << "couldn't read from the file. badly formatted???" << endl;
	  updateFileStream.close();
	  exit(-1);
	}
      
      if(!oneLine.empty())
	{
	  numUpdatesRead++; // count only a non-blank line
	  
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << oneLine;
	  ss >> updateType >> verticalBar >> prefix >> verticalBar
	     >> maskLength >> verticalBar >> nextHop 
	     >> verticalBar >> ASLength;
	  
	  // TODO: check if the line format was correct. if not raise error
	  if(updateType == 'W')
	    RIB.Update(prefix, maskLength, nextHop, ASLength, 0);
	  else if(updateType == 'A')
	    RIB.Update(prefix, maskLength, nextHop, ASLength, 1);
	  else
	    {
	      cout << "Unknown Update Type" << endl;
	      exit(-1);
	    }
	}
      
      if(numUpdatesRead == numUpdToBeRead)
	return numUpdToBeRead;
    }

  resetUpdateFile(); // reset flags and close the file
  return numUpdatesRead;
}

/*******************************************************************
Reads one line at a time from the routing table.
Each line read contains exactly a prefix mask nexthop ASPathLength.
It provides the above four values to the function RIBUpdate().
********************************************************************/

int cRouteProcessor::ReadRoutingTable(cRIB& RIB)
{
  numEntriesReadFromRT = 0;
  ifstream input(RTFilename.c_str());
  
  if(!input.good())
    {
      // file could not be opened for reading
      cout << "Error reading routing table file..." << endl;
      exit(-1);
    }

  // the following loop reads and processes one line in each iteration
  while((input.peek() != EOF) && input.good())
    {
      string oneLine; // the string to be read from the file
  
      // variables to be read from each line of the file
      unsigned int prefix;
      int maskLength;
      unsigned int nextHop;
      int ASLength;

      if(!getline(input, oneLine))
	{
	  cout << "couldn't read from the file. badly formatted???" << endl;
	  input.close();
	  exit(-1);
	}
      
      if(!oneLine.empty())
	{
	  numEntriesReadFromRT++; // count whether its a new or already existing prefix
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << oneLine;
	  ss >> prefix >> maskLength >> nextHop >> ASLength;

	  // TODO: check if the line format was correct. if not raise error

	  RIB.RIBUpdate (prefix, maskLength, nextHop, ASLength, 1);
	}
      
    }

  input.close();
  cout << "Total number of entries: " << numEntriesReadFromRT << endl;
  return 0;
}

int cRouteProcessor::ReadRoutingTable(cRIBdualNH& RIB)
{
  ifstream input(RTFilename.c_str());
  numEntriesReadFromRT = 0;
  
  if(!input.good())
    {
      // file could not be opened for reading
      cout << "Error reading routing table file..." << endl;
      exit(-1);
    }

  // the following loop reads and processes one line in each iteration
  while((input.peek() != EOF) && input.good())
    {
      string oneLine; // the string to be read from the file
  
      // variables to be read from each line of the file
      unsigned int prefix;
      int maskLength;
      unsigned int nextHop;
      int ASLength;

      if(!getline(input, oneLine))
	{
	  cout << "couldn't read from the file. badly formatted???" << endl;
	  input.close();
	  exit(-1);
	}
      
      if(!oneLine.empty())
	{
	  numEntriesReadFromRT++; // count whether its a new or already existing prefix
	  stringstream ss (stringstream::in | stringstream::out);
	  ss << oneLine;
	  ss >> prefix >> maskLength >> nextHop >> ASLength;

	  // TODO: check if the line format was correct. if not raise error

	  RIB.Update(prefix, maskLength, nextHop, ASLength, 1);
	}
      
    }

  input.close();
  cout << "Total number of entries: " << numEntriesReadFromRT << endl;
  return 0;
}
