/***************************************************************************
 *   Copyright (C) 2010 by Ahsan Tariq   *
 *   ahsan@mpi-sws.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <cstring>
#include <cstdlib>
#include <time.h>

#define LEFT 0
#define RIGHT 1

#define ROUTE_SUPPRESSION 1
/******* USER DEFINED HEADERS ********/


#include "FIB.h"
#include "RIB.h"
#include "RIBToFIBDownloader.h"
#include "Algorithms.h"
#include "RouteProcessor.h"
/******* END USER DEFINED HEADERS ***/

#define DEBUG 1

using namespace std;

// the following be set true for ATT result only and NOT for routeviews results
bool ALL_UPDATES_POTENT = true;

int main (int argc, char** argv)
{
  /*
    if (!(argc > 1))
    {
    cout << "Specify configuration file" << endl;
    return 0;
    }
  */
  
  string RoutingTableFile;
  string UpdateFile;
  ifstream ConfigurationFile;
  ofstream SMALTAResultFile;
  ofstream ORTCResultFile;
  
  int TestScenario;
  string Line;
  char *cstr;
  char *tokens;
  int RT = 0;
  int UP = 0;
  int Scen = 0;
  unsigned int ret = 0;
  ConfigurationFile.open("Scenario");
  
  if (!ConfigurationFile)
    {
      cout << "Error ! Could not open configuration file" << endl;
      return 1;
    }
  
  getline (ConfigurationFile, Line);
  
  // Converting to Cstring format
  
  cstr = new char [Line.size() + 1];
  strcpy (cstr, Line.c_str());
  //cout << cstr << "\n";
  // conversion complete

  // Reading the configuration file
  
  tokens = strtok (cstr, " ");
  while (tokens != NULL)
    {
      if (RT || UP || Scen)
	{
	  if (RT)
	    {
	      //cout << "RT";
	      RoutingTableFile += tokens;
	      RT = 0;
	      //cout << "RT";
	    }
	  else if (UP)
	    {
	      UP = 0;
	      //cout << UpdateFile << "\n";
	      UpdateFile += tokens;
	    }
	  else if (Scen)
	    {
	      TestScenario = atoi (tokens);
	      Scen = 0;
	      //cout << "Test Scenario" << TestScenario << "\n";
	      if (TestScenario > 70 || TestScenario < 1)	
		{
		  cout << "Error ! invalid scenario number" << endl;
		  return 1;
		}
	    }
	}
      else if (strcmp (tokens, "RT") == 0)
	{
	  RT = 1;
	  UP = 0;
	  Scen = 0;
	  //cout << tokens << RT << " "<< UP << " " << Scen <<"\n";
	}
      else if (strcmp (tokens, "UP") == 0)
	{
	  //cout << tokens << "\n";
	  RT = 0;
	  UP = 1;
	  Scen = 0;
	}
      else if (strcmp (tokens, "SCEN") == 0)
	{
	  //cout << tokens << "\n";
	  RT = 0;
	  UP = 0;
	  Scen = 1;
	}
      else
	{
	  cout << "Unidentified token " << tokens << " found. Exiting" << endl;
	  return 1;
	}
      tokens = strtok (NULL, " ");
    }
  delete [] cstr;

  if(ALL_UPDATES_POTENT)
    {
      cout << endl
	   << "==================================================" << endl
	   << "* WARNING: All updates potent. Is this ATT data? *" << endl
	   << "==================================================" << endl
	   << endl;
    }
  
  // Reading configuration file complete
  cout << "Running Scenario: " << TestScenario << endl;
  cout << "Routing table: " << RoutingTableFile.c_str() << endl;
  if(UpdateFile.length() > 0)
    cout << "Update File: " << UpdateFile << endl;

  // Remove the path information from the Routing Table file
  // location of last occurence of slash (Unix/Linux) or backslash (windows)
  size_t foundLoc = RoutingTableFile.find_last_of("/\\");
  string RTFileWithoutPath = RoutingTableFile.substr(foundLoc+1);
  
  // the output files be stored in...
  // either the local directory or where the source data is
  // (exactly one of the following be uncommented)
  string RTFilename = RTFileWithoutPath; // stores in local folder
  // string RTFilename = RoutingTableFile; // stores where source data is

  switch (TestScenario)	
    {
    case 13:
      {
	// filename for timing results
	string timingResultsFilename = RTFilename + ".timing";

	// generic stuff
	int numEntriesOT, numUpdates, numEffectiveUpdates;
	clock_t startTime, stopTime;
	double timeForOneShot=0, timeForUpdatingOT=0, timeForUpdatingOTandSMALTA=0, timeForSMALTA=0;
	
	// generic stuff used by all agggregation algorithms
	cRouteProcessor RouteProcessor(RoutingTableFile, UpdateFile);
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cPolicy Policy(1); // policy for picking best route (1 => shorst AS-Path length) 

	// Define the RIBs for each aggregation algorithm
	cRIBdualNH PlainRIB; // the one that we will not attempt to aggregate
	cRIBdualNH ORTCRIB; // RIB that will be aggregated using ORTC/SMALTA

	// Instantiate the algorithm objects
	// These algorithms include one-shot as well as updates
	cORTCUpdate SMALTA(&ORTCRIB, &RIBToFIBDownloader);
	
	// Which algorithm be applied on RIB when an update is incorporated
	// These algorithms DO NOT work until they are enabled explicitly
	// Caution: These algos should be enabled AFTER reading the snapshot
	// because the snapshot is read into the RIB as a sequence of updates
	// A RIB that we don't intend to aggregate does not need an associated algo
	ORTCRIB.SetORTCUpdate(&SMALTA);

	// ****************************
	// Obtain ORTC Timing results
	// ****************************
	cout << "************************" << endl
	     << "Obtaining Timing results" << endl
	     << "************************" << endl;

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(ORTCRIB);
	RouteProcessor.ReadRoutingTable(PlainRIB);
	numEntriesOT = PlainRIB.GetOriginalTable();

	// apply the one-shot algorithm (ORTC) on its RIB
	cout << "Applying the one shot algorithm..." << endl;
	startTime = clock();
	SMALTA.ApplyORTC();
	stopTime = clock();
	timeForOneShot = ((double)(stopTime - startTime))/CLOCKS_PER_SEC;

	// Enable the update algorithm to incorporate the updates
	ORTCRIB.SetORTCUpdateEnabled(1);

	// Read the updates one-by-one and...
	// let them be incorporated as follows:
	// into the original table by the RIB itself.
	// into the aggregated table using the algorithm (if already set and enabled)
	/*	cout << "Applying the updates to Plain RIB..." << endl;
	startTime = clock();
	numUpdates = RouteProcessor.ReadUpdateFile(PlainRIB);
	stopTime = clock();
	timeForUpdatingOT = ((double)(stopTime - startTime))/CLOCKS_PER_SEC;*/

	cout << "Applying the updates to OT and Aggregated RIB..." << endl;
	startTime = clock();
	numUpdates = RouteProcessor.ReadUpdateFile(ORTCRIB);
	stopTime = clock();
	timeForUpdatingOTandSMALTA = ((double)(stopTime - startTime))/CLOCKS_PER_SEC;

	timeForSMALTA = timeForUpdatingOTandSMALTA - timeForUpdatingOT;
	
	numEffectiveUpdates = SMALTA.GetEffectiveUpdates();
	
	// close the snapshot and updates files
	//cout << "Closing the RT and Update Files" << endl;
	//RouteProcessor.ResetRoutingTable();
	//RouteProcessor.ResetUpdateFile();

	// Gather the original table and aggregated table into respective files
	cout << "Writing results to the file (and printing here)..." << endl;
	cout << "#Entries OT: " << numEntriesOT << endl
	     << "#Updates (total): " << numUpdates << endl
	     << "#Updates (potent): " << numEffectiveUpdates << endl
	     << "Time to run one-shot: " << timeForOneShot << endl
	     << "Time to incorporate updates in OT: " << timeForUpdatingOT << endl
	     << "Time to incorporate updates in AT (SMALTA): " << timeForSMALTA << endl;

	ofstream timingResultsFileHandle(timingResultsFilename.c_str());
	timingResultsFileHandle << "#Entries OT: " << numEntriesOT << endl
				<< "#Updates (total): " << numUpdates << endl
				<< "#Updates (potent): " << numEffectiveUpdates << endl
				<< "Time to run one-shot: " << timeForOneShot << endl
				<< "Time to incorporate updates in OT: " << timeForUpdatingOT << endl
				<< "Time to incorporate updates in AT (SMALTA): " << timeForSMALTA << endl;

	timingResultsFileHandle.close();
	
	// ********** Timing Results Finished ************
	break;
      }
    case 11:
      {
	// filenames for one-shot results
	string originalTable = RTFilename + ".orig";
	string originalTableFromORTC = RTFilename + ".fromortc.orig";
	string originalTableFromORTCNR = RTFilename + ".fromortcnr.orig";
	string ORTCTable = RTFilename + ".ortc";
	string ORTCNRTable = RTFilename + ".ortcnr";
	string L1Table = RTFilename + ".l1";
	string L2Table = RTFilename + ".l2";

	// filenames for results after incorporating the updates
	// final results represent AFTER ALL updates have been incorporated
	string updatedOriginalTable = RTFilename + ".upd.orig";
	string updatedOriginalTableFromORTC = RTFilename + ".upd.fromortc.orig";
	string updatedOriginalTableFromORTCNR = RTFilename + ".upd.fromortcnr.orig";
	string updatedORTCTable = RTFilename + ".smalta";
	string updatedORTCNRTable = RTFilename + ".upd.ortcnr";
	string updatedL1Table = RTFilename + ".upd.l1";
	string updatedL2Table = RTFilename + ".upd.l2";

	// generic stuff used by all agggregation algorithms
	cRouteProcessor RouteProcessor(RoutingTableFile, UpdateFile);
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cPolicy Policy(1); // policy for picking best route (1 => shorst AS-Path length) 

	// Define the RIBs for each aggregation algorithm
	cRIB L1RIB(&Policy, 0); // RIB used for L1
	cRIB L2RIB(&Policy, 0); // RIB used for L2
	cRIB ORTCNRRIB(&Policy, 0); // RIB used for ORTC-NR
	cRIBdualNH ORTCRIB; // RIB used for ORTC

	// Instantiate the algorithm objects
	// These algorithms include one-shot as well as updates
	cORTCUpdate SMALTA(&ORTCRIB, &RIBToFIBDownloader);
	cSMALTA ORTCNRAlgo(&ORTCNRRIB, &RIBToFIBDownloader); // ORTC-NR with Updates
	cZhangAgg ZhangAgg1(&L1RIB, &RIBToFIBDownloader, 1);
	cZhangAgg ZhangAgg2(&L2RIB, &RIBToFIBDownloader, 2);
	
	// Which algorithm be applied on RIB when an update is incorporated
	// These algorithms DO NOT work until they are enabled explicitly
	// Caution: These algos should be enabled AFTER reading the snapshot
	// because the snapshot is read into the RIB as a sequence of updates
	L1RIB.SetZhangAgg(&ZhangAgg1);
	L2RIB.SetZhangAgg(&ZhangAgg2);
	ORTCNRRIB.SetSMALTA(&ORTCNRAlgo);
	ORTCRIB.SetORTCUpdate(&SMALTA);
	
	// setting the routing table file and the update file
	// simply opnes the files for reading
	//	RouteProcessor.SetRoutingTable (&RoutingTableFile);
	//	RouteProcessor.SetUpdateFile (&UpdateFile);

	// **********************
	// Obtain Level 1 results
	// **********************
	cout << "*************************" << endl
	     << "Obtaining Level 1 results" << endl
	     << "*************************" << endl;

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(L1RIB);

	// apply the one-shot algorithm (Level 1) on its RIB
	cout << "Applying the one shot algorithm..." << endl;
	ZhangAgg1.Apply();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	L1RIB.GetOriginalTable(&originalTable);
	L1RIB.GetAggregatedTable(&L1Table);

	// Enable the update algorithm to incorporate the updates
	L1RIB.SetZhangAggEnabled(1);

	// Read the updates one-by-one and...
	// let them be incorporated as follows:
	// into the original table by the RIB itself.
	// into the aggregated table using the algorithm (already set and enabled)
	cout << "Applying the updates..." << endl;
	RouteProcessor.ReadUpdateFile(L1RIB);

	// close the snapshot and updates files
	//cout << "Closing the RT and Update Files" << endl;
	//RouteProcessor.ResetRoutingTable();
	//RouteProcessor.ResetUpdateFile();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	L1RIB.GetOriginalTable(&updatedOriginalTable);
	L1RIB.GetAggregatedTable(&updatedL1Table);	
	
	// ********** Level 1 Results Finished ************

	// **********************
	// Obtain ORTC-NR results
	// **********************
	/*cout << "*************************" << endl
	     << "Obtaining ORTC-NR results" << endl
	     << "*************************" << endl;

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(ORTCNRRIB);

	// apply the one-shot algorithm (ORTC-NR) on its RIB
	cout << "Applying the one shot algorithm..." << endl;
	ORTCNRAlgo.ApplyORTCNR();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	ORTCNRRIB.GetOriginalTable(&originalTableFromORTCNR);
	ORTCNRRIB.GetAggregatedTable(&ORTCNRTable);

	// Enable the update algorithm to incorporate the updates
	ORTCNRRIB.SetSMALTAEnabled(1);

	// Read the updates one-by-one and...
	// let them be incorporated as follows:
	// into the original table by the RIB itself.
	// into the aggregated table using the algorithm (already set and enabled)
	cout << "Applying the updates..." << endl;
	RouteProcessor.ReadUpdateFile(ORTCNRRIB);

	// close the snapshot and updates files
	//cout << "Closing the RT and Update Files" << endl;
	//RouteProcessor.ResetRoutingTable();
	//RouteProcessor.ResetUpdateFile();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	ORTCNRRIB.GetOriginalTable(&updatedOriginalTableFromORTCNR);
	ORTCNRRIB.GetAggregatedTable(&updatedORTCNRTable);	
	*/
	// ********** ORTC-NR Results Finished ************

	// **********************
	// Obtain Level 2 results
	// **********************
	cout << "*************************" << endl
	     << "Obtaining Level 2 results" << endl
	     << "*************************" << endl;

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(L2RIB);

	// apply the one-shot algorithm (Level 2) on its RIB
	cout << "Applying the one shot algorithm..." << endl;
	ZhangAgg2.Apply();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	//L2RIB.GetOriginalTable(&originalTable);
	L2RIB.GetAggregatedTable(&L2Table);

	// Enable the update algorithm to incorporate the updates
	L2RIB.SetZhangAggEnabled(1);

	// Read the updates one-by-one and...
	// let them be incorporated as follows:
	// into the original table by the RIB itself.
	// into the aggregated table using the algorithm (already set and enabled)
	cout << "Applying the updates..." << endl;
	RouteProcessor.ReadUpdateFile(L2RIB);

	// close the snapshot and updates files
	//cout << "Closing the RT and Update Files" << endl;
	//RouteProcessor.ResetRoutingTable();
	//RouteProcessor.ResetUpdateFile();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	//L2RIB.GetOriginalTable(&updatedOriginalTable);
	L2RIB.GetAggregatedTable(&updatedL2Table);
		
	// ********** Level 2 Results Finished ************

	// **********************
	// Obtain ORTC results
	// **********************
	cout << "**********************" << endl
	     << "Obtaining ORTC results" << endl
	     << "**********************" << endl;

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(ORTCRIB);

	// apply the one-shot algorithm (ORTC) on its RIB
	cout << "Applying the one shot algorithm..." << endl;
	SMALTA.ApplyORTC();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	ORTCRIB.GetOriginalTable(&originalTableFromORTC);
	ORTCRIB.GetAggregatedTable(&ORTCTable);

	// Enable the update algorithm to incorporate the updates
	ORTCRIB.SetORTCUpdateEnabled(1);

	// Read the updates one-by-one and...
	// let them be incorporated as follows:
	// into the original table by the RIB itself.
	// into the aggregated table using the algorithm (already set and enabled)
	cout << "Applying the updates..." << endl;
	RouteProcessor.ReadUpdateFile(ORTCRIB);

	// close the snapshot and updates files
	//cout << "Closing the RT and Update Files" << endl;
	//RouteProcessor.ResetRoutingTable();
	//RouteProcessor.ResetUpdateFile();

	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	ORTCRIB.GetOriginalTable(&updatedOriginalTableFromORTC);
	ORTCRIB.GetAggregatedTable(&updatedORTCTable);
			
	// ********** ORTC Results Finished ************
	break;
      }

    case 12:
      {
	string ORTCStatsFilename = RTFilename + ".prog.ortc.stats";
	string SMALTAStatsFilename = RTFilename + ".prog.smalta.stats";
	string progOriginalTableFromORTC = RTFilename + ".prog.fromortc.orig";
	string progOriginalTableFromSMALTA = RTFilename + ".prog.fromsmalta.orig";
	string progPerPrefixResults = RTFilename + ".prog.perprefix-smalta.stats";
	string progORTCTable = RTFilename + ".prog.ortc";
	string progSMALTATable = RTFilename + ".prog.smalta";
	
	// generic stuff
	int numEntriesAT, numEntriesOT;
	unsigned int rib_to_fib_downloads;

	// generic stuff
	cRouteProcessor RouteProcessor(RoutingTableFile, UpdateFile);
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cPolicy Policy(1); // policy for picking best route (1 => shorst AS-Path length)
	
	//number of updates to be read in one go
	int numUpdToBeRead = 10000;
	
	// Define the RIBs for each aggregation algorithm
	cRIBdualNH ORTCRIB; // RIB used for ORTC results
	cRIBdualNH SMALTARIB; // RIB used for SMALTA results

	// Instantiate the algorithm object
	// The algorithm includes one-shot as well as updates
	// one algorithm works with both ORTC and SMALTA progressive results
	cORTCUpdate SMALTA1(&ORTCRIB, &RIBToFIBDownloader);
	cORTCUpdate SMALTA2(&SMALTARIB, &RIBToFIBDownloader);
	
	// Which algorithm be applied on RIB when an update is incorporated
	// These algorithms DO NOT work until they are enabled explicitly
	// Caution: These algos should be enabled AFTER reading the snapshot
	// because the snapshot is read into the RIB as a sequence of updates
	ORTCRIB.SetORTCUpdate(&SMALTA1);
	SMALTARIB.SetORTCUpdate(&SMALTA2);

	// *******************************
	// Obtain progressive ORTC results
	// TODO -- what these results are?
	// *******************************
	cout << "**********************************" << endl
	     << "Obtaining progressive ORTC results" << endl
	     << "**********************************" << endl;

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(ORTCRIB);

	// apply the one-shot algorithm (ORTC) on its RIB
	// and gather some stats about it
	cout << "Applying the one shot algorithm..." << endl;
	numEntriesAT = SMALTA1.ApplyORTC();
	rib_to_fib_downloads = SMALTA1.GetORTCRIBtoFIBdl();

	ofstream output(ORTCStatsFilename.c_str());
	output << "#Entries after one-shot: " << numEntriesAT << endl
	       << "#RIB-to-FIB downloads: " << rib_to_fib_downloads << endl;
	
	// Enable the update algorithm to incorporate the updates
	ORTCRIB.SetORTCUpdateEnabled(1);

	// Gather the stats
	output << "Progressive Stats" << endl
	       << "=================" << endl
	       << "#new-upds\t#eff.-upds\t#Entries(OT)\t#TheorMinEntries\t#RIB2FIB-DLs" << endl;
	while(true)
	  {
	    int numUpdatesRead, numEffectiveUpdates;
	    numUpdatesRead = RouteProcessor.ReadUpdateFile(ORTCRIB, numUpdToBeRead);
	    
	    numEffectiveUpdates = SMALTA1.GetEffectiveUpdates();
	    numEntriesOT = ORTCRIB.GetOriginalTable(); // called w/o arguments
	    numEntriesAT = SMALTA1.ApplyORTC();
	    rib_to_fib_downloads = SMALTA1.GetORTCRIBtoFIBdl();
	    output << numUpdatesRead << "\t" << numEffectiveUpdates << "\t"
		   << numEntriesOT << "\t" << numEntriesAT << "\t" 
		   << rib_to_fib_downloads << endl;

	    if(numUpdatesRead < numUpdToBeRead)
	      break;
	  }
	
	output.close();
	
	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	ORTCRIB.GetOriginalTable(&progOriginalTableFromORTC);
	ORTCRIB.GetAggregatedTable(&progORTCTable);

	// *********************************
	// Obtain progressive SMALTA results
	// TODO -- what these results are?
	// *********************************
	cout << "************************************" << endl
	     << "Obtaining progressive SMALTA results" << endl
	     << "************************************" << endl;

	// first attach the SMALTARIB with the update algorithm
	// which may still have ORTCRIB object attached
	// this associateNewRIB also needs to reset other stats in the algorithms
	// holding it for now -- and using a different instance of algorithm
	// SMALTA1.associateNewRIB(&SMALTARIB);

	// reading the routing data into the RIB -- the snapshot
	cout << "Reading the Routing table file..." << endl;
	RouteProcessor.ReadRoutingTable(SMALTARIB);

	// apply the one-shot algorithm (ORTC) on its RIB
	// and gather some stats about it
	cout << "Applying the one shot algorithm..." << endl;
	numEntriesAT = SMALTA2.ApplyORTC();
	rib_to_fib_downloads = SMALTA2.GetORTCRIBtoFIBdl();

	// debug information
	cout << "#Entries AT after One-shot: " << SMALTARIB.GetAggregatedTable() << endl;

	// open file for writing per prefix results
	// note: smalta results are written using same file handle as ORTC)
	ofstream perPrefixOutput(progPerPrefixResults.c_str());
	output.open(SMALTAStatsFilename.c_str());
	output << "#Entries after one-shot: " << numEntriesAT << endl
	       << "#RIB-to-FIB downloads: " << rib_to_fib_downloads << endl;

	// Enable the update algorithm to incorporate the updates
	SMALTARIB.SetORTCUpdateEnabled(1);

	// Gather the stats
	output << "Progressive Stats" << endl
	       << "=================" << endl
	       << "#new-upds\t#eff.-upds\t#entriesOT\t#entriesAT\tRIB2FIB-DLs" << endl;
	while(true)
	  {
	    int numUpdatesRead, numEffectiveUpdates;
	    numUpdatesRead = RouteProcessor.ReadUpdateFile(SMALTARIB, numUpdToBeRead);
	    
	    numEffectiveUpdates = SMALTA2.GetEffectiveUpdates();
	    numEntriesOT = SMALTARIB.GetOriginalTable(); // called w/o arguments
	    numEntriesAT = SMALTARIB.GetAggregatedTable(); // called w/o arguments
	    rib_to_fib_downloads = SMALTA2.GetSMALTARIBtoFIBdl();
	    output << numUpdatesRead << "\t" << numEffectiveUpdates << "\t"
		   << numEntriesOT << "\t" << numEntriesAT << "\t" 
		   << rib_to_fib_downloads << endl;

	    // write per prefix results to its own file
	    unsigned int *effectiveUpdatesPerPrefix = SMALTA2.GetEffectiveUpdatesByPrefix();
	    unsigned int *nodesChangedPerPrefix = SMALTA2.GetNodesChangedByPrefix();
	    perPrefixOutput << "Prefix Length" << " " <<"#updates" 
			    << " " << "#nodes changed" << endl;
	    for(int i=0; i < 33; i++)
	      perPrefixOutput << i << " " << effectiveUpdatesPerPrefix[i] << " "
			      << nodesChangedPerPrefix[i] << endl;

	    if(numUpdatesRead < numUpdToBeRead)
	      break;
	  }
	
	perPrefixOutput.close();
	output.close();
	
	// Gather the original table and aggregated table into respective files
	cout << "Writing output: Original Table and Aggregated Table" << endl;
	SMALTARIB.GetOriginalTable(&progOriginalTableFromSMALTA);
	SMALTARIB.GetAggregatedTable(&progSMALTATable);

	break;
      }

      /*
    case 1:  // Apply ORTC only
      {
	string OutputFile;
	string ORTCTable;
	string SMALTATable;
	OutputFile += RoutingTableFile;
	OutputFile += ".orig";
	ORTCTable += RoutingTableFile;
	ORTCTable += ".l1";
	SMALTATable += RoutingTableFile;
	SMALTATable += ".upd.l1";
	cout << "OutPut File: " << OutputFile << "\n";
	cPolicy Policy(1);
	cRIB RIB(&Policy, 0);
	cRouteProcessor RouteProcessor;
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cSMALTA SMALTA (&RIB, &RIBToFIBDownloader);
	cZhangAgg ZhangAgg (&RIB, &RIBToFIBDownloader, 1);
	RIB.SetSMALTA (&SMALTA);
	RIB.SetZhangAgg (&ZhangAgg);
	if (RouteProcessor.SetRoutingTable (&RoutingTableFile))
	  {
	    cout << "Route Processor Routing Table not Set\n";
	    return 1;
	  }
	RouteProcessor.SetUpdateFile (&UpdateFile);
	RouteProcessor.ReadRoutingTable(&RIB);
	cout << "Routing table Read\n";
	
	//RIB.GetOriginalTable (&OutputFile);
	RIB.SetSMALTAEnabled (1);
	//cout << "Applying level 1\n";
	start = clock ();
	ZhangAgg.Apply();
	RIB.GetOriginalTable (&OutputFile);
	stop = clock ();
	one_shot = ((double)(stop - start))/CLOCKS_PER_SEC;
	RIB.GetAggregatedTable (&ORTCTable);
	start = clock ();
	ret = RouteProcessor.ReadUpdateFile (-1, &RIB);
	stop = clock ();
	cout << "average time to incorprate " << ret << " updates = " << (stop - start)/((double) CLOCKS_PER_SEC)/ret << "\n";
	cout << "Update File read\n";
	RIB.GetAggregatedTable (&SMALTATable);
	cout << "total time for one shot = " << one_shot << "\n";
	cout << "Nodes Changed Level1 = " << ZhangAgg.GetNodesChangedLevel1() << "\n";
	cout << "Average Nodes changed Level 1 = " << ((double)ZhangAgg.GetNodesChangedLevel1())/ret << "\n";
	cout << "Nodes Changed Level2 = " << ZhangAgg.GetNodesChangedLevel2() << "\n";
	cout << "Average Nodes changed Level 2 = " << ((double)ZhangAgg.GetNodesChangedLevel2())/ret << "\n";
	cout << "Nodes Changed SMALTA = " << SMALTA.GetNodesChanged() << "\n";
	cout << "Average Nodes changed SMALTA = " << ((double)SMALTA.GetNodesChanged())/ret << "\n";
	//RIB.SetSMALTAEnabled (1);		// enable SMALTA
	//ORTC.ApplyORTC();
	//ORTC.GetORTCTable (&ORTCTable);
	
	//RouteProcessor.ReadUpdateFile (-1, &RIB);
	//cout << "Update File read\n";
	//RIB.GetOriginalTable (&OutputFile);
	//ORTC.GetORTCTable (&SMALTATable);
	
	FIB.OutputFIB (&OutputFile);
	break;
      }
    case 2:		// Apply ORTC then all updates by SMALTA
      {
	string ORTCTable = "ORTC";
	string Original = "Original";
	string OriginalBUP = "OriginalBUP";
	string UpdateTable = "Update";
	string snap = "snap";
	int entriesOneShot;
	int entriesAftUpd;						
	cPolicy Policy(1);
	cRIB RIB(&Policy, 0);
	cRouteProcessor RouteProcessor;
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cSMALTA SMALTA (&RIB, &RIBToFIBDownloader);
	cZhangAgg ZhangAgg (&RIB, &RIBToFIBDownloader, 1);
	RIB.SetSMALTA (&SMALTA);
	RIB.SetZhangAgg (&ZhangAgg);
	if (RouteProcessor.SetRoutingTable (&RoutingTableFile))
	  {
	    cout << "Route Processor Routing Table not Set\n";
	    return 1;
	  }
	RouteProcessor.SetUpdateFile (&UpdateFile);
	cout << "Reading Routing Table" << endl;
	RouteProcessor.ReadRoutingTable(&RIB);
	cout << "Routing table Read\n";
	RIB.SetZhangAggEnabled(1);
	//RIB.SetSMALTAEnabled (1);
	//entriesOneShot = SMALTA.ApplyORTCNR();	
	ZhangAgg.Apply();
	RIB.GetOriginalTable (&OriginalBUP);
	RIB.GetAggregatedTable (&ORTCTable);
	ret = RouteProcessor.ReadUpdateFile (-1,&RIB);
	RIB.GetOriginalTable (&Original);
	entriesAftUpd = RIB.GetAggregatedTable (&UpdateTable);
	cout << "Number of Entries in Agg. Table before updates = " << entriesOneShot << endl;
	cout << "Number of Entries in Agg. Table After updates (SMALTA) = " << entriesAftUpd << endl;
	cout << "Total Updates = " << ret << endl;
	break;
      }
    case 3:		// level 1 one shot
      {
	string AggTable = "OneShotAggregated";
	string OT = "Original";
	int entriesOneShot;						
	cPolicy Policy(1);
	cRIB RIB(&Policy, 0);
	cRouteProcessor RouteProcessor;
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cZhangAgg ZhangAgg (&RIB, &RIBToFIBDownloader, 1);
	RIB.SetZhangAgg (&ZhangAgg);
	if (RouteProcessor.SetRoutingTable (&RoutingTableFile))	{
	  cout << "Route Processor Routing Table not Set\n";
	  return 1;
	}
	RouteProcessor.ReadRoutingTable(&RIB);
	cout << "Routing table Read\n";

	RIB.SetZhangAggEnabled(1);
	ZhangAgg.Apply();
	RIB.GetOriginalTable (&OT);
	entriesOneShot = RIB.GetAggregatedTable (&AggTable);
	cout << "Number of Entries in Agg. Table After One Shot level 1 = " << entriesOneShot<< endl;					     
	break;
      }
    case 4:		// level 2 one shot
      {
	string AggTable = "OneShotAggregated";
	int entriesOneShot;						
	cPolicy Policy(1);
	cRIB RIB(&Policy, 0);
	cRouteProcessor RouteProcessor;
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cZhangAgg ZhangAgg (&RIB, &RIBToFIBDownloader, 2);
	RIB.SetZhangAgg (&ZhangAgg);
	if (RouteProcessor.SetRoutingTable (&RoutingTableFile))	{
	  cout << "Route Processor Routing Table not Set\n";
	  return 1;
	}
	RouteProcessor.ReadRoutingTable(&RIB);
	cout << "Routing table Read\n";

	RIB.SetZhangAggEnabled(1);
	ZhangAgg.Apply();
	entriesOneShot = RIB.GetAggregatedTable (&AggTable);
	cout << "Number of Entries in Agg. Table After One Shot level 2 = " << entriesOneShot << endl;					     
	break;
      }
		
    case 5:
      {
	string ORTCTable = "ORTC";
	string OriginalTable = "OriginalTable";
	int entriesOneShot;
	//int entriesAftUpd;						
	cPolicy Policy(1);
	cRIB RIB(&Policy, 0);
	cRIBdualNH RIBdualNH;
	cRouteProcessor RouteProcessor;
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cORTCUpdate ORTCUpdate(&RIBdualNH, &RIBToFIBDownloader);
	RouteProcessor.SetRoutingTable (&RoutingTableFile);
	RouteProcessor.SetUpdateFile (&UpdateFile);
	RIBdualNH.SetORTCUpdate (&ORTCUpdate);
	RIBdualNH.SetORTCUpdateEnabled (0);
	cout << "Reading Routing Table\n";
	RouteProcessor.ReadRoutingTable(&RIBdualNH);
	cout << "Routing table Read\n";
	ORTCUpdate.ApplyORTC();
	RIBdualNH.SetORTCUpdateEnabled (1);
	RouteProcessor.ReadUpdateFile (-1, &RIBdualNH);
	RIBdualNH.GetOriginalTable(&OriginalTable);
	entriesOneShot = RIBdualNH.GetAggregatedTable (&ORTCTable);
			
	cout << "Number of Entries in Agg. Table After oneshot ORTC = " << entriesOneShot << endl;
	break;
      }
		
    case 6:
      {
	string ORTCNRTable = "ORTCNR";
	string OriginalTable = "OriginalTable";
	int entriesOneShot;
	cPolicy Policy(1);
	cRIB RIB(&Policy, 0);
	cRouteProcessor RouteProcessor;
	cFIB FIB;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cSMALTA SMALTA (&RIB, &RIBToFIBDownloader);
	RIB.SetSMALTA (&SMALTA);;
	RouteProcessor.SetRoutingTable (&RoutingTableFile);
	RouteProcessor.SetUpdateFile (&UpdateFile);
	RouteProcessor.ReadRoutingTable (&RIB);
	entriesOneShot = SMALTA.ApplyORTCNR();
	RIB.SetSMALTAEnabled (1);
	cout << "entries = " << entriesOneShot << endl;
	ret = RouteProcessor.ReadUpdateFile (100000, &RIB);
	entriesOneShot = SMALTA.ApplyORTCNR();
	RIB.GetOriginalTable (&OriginalTable);
	cout << "entries = " << entriesOneShot << endl;

      }
      
	case 7:
	{
	string Statistics += RoutingTableFile;
	string Statistics += ".stat";
	cPolicy Policy(1);
	cRIB RIB1(&Policy, 0);
	cRIB RIB2(&Policy, 0);
	cRouteProcessor RouteProcessor1, RouteProcessor2;
	cRIBToFIBDownloader RIBToFIBDownloader;
	cSMALTA SMALTA1(&RIB1, &RIBToFIBDownloader);
	cSMALTA SMALTA2(&RIB2, &RIBToFIBDownloader);
	RIB1.SetSMALTA (&SMALTA1);
	RIB2.SetSMALTA (&SMALTA2);
	RouteProcessor1.SetRoutingTable (&RoutingTableFile);
	RouteProcessor2.SetRoutingTable (&RoutingTableFile);
	RouteProcessor1.SetUpdateFile (&UpdateFile);
	RouteProcessor2.SetUpdateFile (&UpdateFile);
			
	}
	case 8:
	{
	// write the announcemens from each Nexthop to a seperate file
	// also output data on the number of nexthops per prefix
	ifstream RT.open(RoutingTableFile);
	string str_LineRT;
	char* c_LineRT, *tokens;
	vector <unsigned int> NHList;
	vector <string> OutFiles;
	vector <ofstream> ofiles;
	unsigned int Prefix;
	int mask;
	unsigned int NextHop;
	int ASLength;
	int w = -1;
	while (getline(RoutingTable, str_LineRT))	{
				
	c_LineRT = new char [str_LineRT.size() + 1];
	strcpy (c_LineRT, str_LineRT.c_str());
	// get Prefix
	tokens = strtok (c_LineRT, " ");
	//cout << tokens << endl;
	Prefix = atoi (tokens);
			
	// get mask
	tokens = strtok (NULL, " ");
	//cout << tokens << endl;
	mask = atoi (tokens);
			
	// get NextHop
	tokens = strtok (NULL, " ");
	//cout << tokens << endl;
	NextHop = atoi (tokens);
	for (int i = 0; i < NHList.size(); i++)	{
	if (NHList[i] == NextHop)
	w = i;
	break;
	}
	if (w == -1)	{
	w = NHList.size();
	NHList.push_back(NextHop);
	char *cstr = new char[11];
	itoa (NextHop, cstr, 10);
	string s += cstr;
	OutFiles.push_back (s);
	ofstream d;
	ofiles.push_back(d);
	ofiles[w].open(Outfiles[w]);
	}
				
	// get ASLength
	tokens = strtok (NULL, " ");
	//cout << tokens << endl;
	ASLength = atoi (tokens);
				
	delete [] c_LineRT;
		
		
	}
	}*/
    }

  return ret;	
}

